'''Wrapper for /usr/include/EGL/egl

Generated with:
wrap.py -o lib_egl.py /usr/include/EGL/egl.h

Do not modify this file.
'''

__docformat__ =  'restructuredtext'
__version__ = '$Id$'

import ctypes
from ctypes import *

import pyglet.lib

_lib = pyglet.lib.load_library('EGL')

_int_types = (c_int16, c_int32)
if hasattr(ctypes, 'c_int64'):
    # Some builds of ctypes apparently do not have c_int64
    # defined; it's a pretty good bet that these builds do not
    # have 64-bit pointers.
    _int_types += (ctypes.c_int64,)
for t in _int_types:
    if sizeof(t) == sizeof(c_size_t):
        c_ptrdiff_t = t

class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]



__egl_h_ = 1 	# /usr/include/EGL/egl.h:2
EGL_EGL_PROTOTYPES = 1 	# /usr/include/EGL/egl.h:42
EGL_VERSION_1_0 = 1 	# /usr/include/EGL/egl.h:57
EGLBoolean = c_uint 	# /usr/include/EGL/egl.h:58
EGLDisplay = POINTER(None) 	# /usr/include/EGL/egl.h:59
EGLConfig = POINTER(None) 	# /usr/include/EGL/egl.h:62
EGLSurface = POINTER(None) 	# /usr/include/EGL/egl.h:63
EGLContext = POINTER(None) 	# /usr/include/EGL/egl.h:64
__eglMustCastToProperFunctionPointerType = CFUNCTYPE(None) 	# /usr/include/EGL/egl.h:65
EGL_ALPHA_SIZE = 12321 	# /usr/include/EGL/egl.h:66
EGL_BAD_ACCESS = 12290 	# /usr/include/EGL/egl.h:67
EGL_BAD_ALLOC = 12291 	# /usr/include/EGL/egl.h:68
EGL_BAD_ATTRIBUTE = 12292 	# /usr/include/EGL/egl.h:69
EGL_BAD_CONFIG = 12293 	# /usr/include/EGL/egl.h:70
EGL_BAD_CONTEXT = 12294 	# /usr/include/EGL/egl.h:71
EGL_BAD_CURRENT_SURFACE = 12295 	# /usr/include/EGL/egl.h:72
EGL_BAD_DISPLAY = 12296 	# /usr/include/EGL/egl.h:73
EGL_BAD_MATCH = 12297 	# /usr/include/EGL/egl.h:74
EGL_BAD_NATIVE_PIXMAP = 12298 	# /usr/include/EGL/egl.h:75
EGL_BAD_NATIVE_WINDOW = 12299 	# /usr/include/EGL/egl.h:76
EGL_BAD_PARAMETER = 12300 	# /usr/include/EGL/egl.h:77
EGL_BAD_SURFACE = 12301 	# /usr/include/EGL/egl.h:78
EGL_BLUE_SIZE = 12322 	# /usr/include/EGL/egl.h:79
EGL_BUFFER_SIZE = 12320 	# /usr/include/EGL/egl.h:80
EGL_CONFIG_CAVEAT = 12327 	# /usr/include/EGL/egl.h:81
EGL_CONFIG_ID = 12328 	# /usr/include/EGL/egl.h:82
EGL_CORE_NATIVE_ENGINE = 12379 	# /usr/include/EGL/egl.h:83
EGL_DEPTH_SIZE = 12325 	# /usr/include/EGL/egl.h:84
EGL_DRAW = 12377 	# /usr/include/EGL/egl.h:86
EGL_EXTENSIONS = 12373 	# /usr/include/EGL/egl.h:87
EGL_FALSE = 0 	# /usr/include/EGL/egl.h:88
EGL_GREEN_SIZE = 12323 	# /usr/include/EGL/egl.h:89
EGL_HEIGHT = 12374 	# /usr/include/EGL/egl.h:90
EGL_LARGEST_PBUFFER = 12376 	# /usr/include/EGL/egl.h:91
EGL_LEVEL = 12329 	# /usr/include/EGL/egl.h:92
EGL_MAX_PBUFFER_HEIGHT = 12330 	# /usr/include/EGL/egl.h:93
EGL_MAX_PBUFFER_PIXELS = 12331 	# /usr/include/EGL/egl.h:94
EGL_MAX_PBUFFER_WIDTH = 12332 	# /usr/include/EGL/egl.h:95
EGL_NATIVE_RENDERABLE = 12333 	# /usr/include/EGL/egl.h:96
EGL_NATIVE_VISUAL_ID = 12334 	# /usr/include/EGL/egl.h:97
EGL_NATIVE_VISUAL_TYPE = 12335 	# /usr/include/EGL/egl.h:98
EGL_NONE = 12344 	# /usr/include/EGL/egl.h:99
EGL_NON_CONFORMANT_CONFIG = 12369 	# /usr/include/EGL/egl.h:100
EGL_NOT_INITIALIZED = 12289 	# /usr/include/EGL/egl.h:101
EGL_PBUFFER_BIT = 1 	# /usr/include/EGL/egl.h:105
EGL_PIXMAP_BIT = 2 	# /usr/include/EGL/egl.h:106
EGL_READ = 12378 	# /usr/include/EGL/egl.h:107
EGL_RED_SIZE = 12324 	# /usr/include/EGL/egl.h:108
EGL_SAMPLES = 12337 	# /usr/include/EGL/egl.h:109
EGL_SAMPLE_BUFFERS = 12338 	# /usr/include/EGL/egl.h:110
EGL_SLOW_CONFIG = 12368 	# /usr/include/EGL/egl.h:111
EGL_STENCIL_SIZE = 12326 	# /usr/include/EGL/egl.h:112
EGL_SUCCESS = 12288 	# /usr/include/EGL/egl.h:113
EGL_SURFACE_TYPE = 12339 	# /usr/include/EGL/egl.h:114
EGL_TRANSPARENT_BLUE_VALUE = 12341 	# /usr/include/EGL/egl.h:115
EGL_TRANSPARENT_GREEN_VALUE = 12342 	# /usr/include/EGL/egl.h:116
EGL_TRANSPARENT_RED_VALUE = 12343 	# /usr/include/EGL/egl.h:117
EGL_TRANSPARENT_RGB = 12370 	# /usr/include/EGL/egl.h:118
EGL_TRANSPARENT_TYPE = 12340 	# /usr/include/EGL/egl.h:119
EGL_TRUE = 1 	# /usr/include/EGL/egl.h:120
EGL_VENDOR = 12371 	# /usr/include/EGL/egl.h:121
EGL_VERSION = 12372 	# /usr/include/EGL/egl.h:122
EGL_WIDTH = 12375 	# /usr/include/EGL/egl.h:123
EGL_WINDOW_BIT = 4 	# /usr/include/EGL/egl.h:124
khronos_int32_t = c_int32 	# /usr/include/KHR/khrplatform.h:150
EGLint = khronos_int32_t 	# /usr/include/EGL/eglplatform.h:166
PFNEGLCHOOSECONFIGPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, POINTER(EGLint), POINTER(EGLConfig), EGLint, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:125
XID = c_ulong 	# /usr/include/X11/X.h:66
Pixmap = XID 	# /usr/include/X11/X.h:102
EGLNativePixmapType = Pixmap 	# /usr/include/EGL/eglplatform.h:132
PFNEGLCOPYBUFFERSPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLNativePixmapType) 	# /usr/include/EGL/egl.h:126
PFNEGLCREATECONTEXTPROC = CFUNCTYPE(EGLContext, EGLDisplay, EGLConfig, EGLContext, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:127
PFNEGLCREATEPBUFFERSURFACEPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLConfig, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:128
PFNEGLCREATEPIXMAPSURFACEPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLConfig, EGLNativePixmapType, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:129
Window = XID 	# /usr/include/X11/X.h:96
EGLNativeWindowType = Window 	# /usr/include/EGL/eglplatform.h:133
PFNEGLCREATEWINDOWSURFACEPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLConfig, EGLNativeWindowType, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:130
PFNEGLDESTROYCONTEXTPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLContext) 	# /usr/include/EGL/egl.h:131
PFNEGLDESTROYSURFACEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface) 	# /usr/include/EGL/egl.h:132
PFNEGLGETCONFIGATTRIBPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLConfig, EGLint, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:133
PFNEGLGETCONFIGSPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, POINTER(EGLConfig), EGLint, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:134
PFNEGLGETCURRENTDISPLAYPROC = CFUNCTYPE(EGLDisplay) 	# /usr/include/EGL/egl.h:135
PFNEGLGETCURRENTSURFACEPROC = CFUNCTYPE(EGLSurface, EGLint) 	# /usr/include/EGL/egl.h:136
class struct__XDisplay(Structure):
    __slots__ = [
    ]
struct__XDisplay._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XDisplay(Structure):
    __slots__ = [
    ]
struct__XDisplay._fields_ = [
    ('_opaque_struct', c_int)
]

Display = struct__XDisplay 	# /usr/include/X11/Xlib.h:487
EGLNativeDisplayType = POINTER(Display) 	# /usr/include/EGL/eglplatform.h:131
PFNEGLGETDISPLAYPROC = CFUNCTYPE(EGLDisplay, EGLNativeDisplayType) 	# /usr/include/EGL/egl.h:137
PFNEGLGETERRORPROC = CFUNCTYPE(EGLint) 	# /usr/include/EGL/egl.h:138
PFNEGLGETPROCADDRESSPROC = CFUNCTYPE(__eglMustCastToProperFunctionPointerType, c_char_p) 	# /usr/include/EGL/egl.h:139
PFNEGLINITIALIZEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, POINTER(EGLint), POINTER(EGLint)) 	# /usr/include/EGL/egl.h:140
PFNEGLMAKECURRENTPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLSurface, EGLContext) 	# /usr/include/EGL/egl.h:141
PFNEGLQUERYCONTEXTPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLContext, EGLint, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:142
PFNEGLQUERYSTRINGPROC = CFUNCTYPE(c_char_p, EGLDisplay, EGLint) 	# /usr/include/EGL/egl.h:143
PFNEGLQUERYSURFACEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLint, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:144
PFNEGLSWAPBUFFERSPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface) 	# /usr/include/EGL/egl.h:145
PFNEGLTERMINATEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay) 	# /usr/include/EGL/egl.h:146
PFNEGLWAITGLPROC = CFUNCTYPE(EGLBoolean) 	# /usr/include/EGL/egl.h:147
PFNEGLWAITNATIVEPROC = CFUNCTYPE(EGLBoolean, EGLint) 	# /usr/include/EGL/egl.h:148
# /usr/include/EGL/egl.h:150
eglChooseConfig = _lib.eglChooseConfig
eglChooseConfig.restype = EGLBoolean
eglChooseConfig.argtypes = [EGLDisplay, POINTER(EGLint), POINTER(EGLConfig), EGLint, POINTER(EGLint)]

# /usr/include/EGL/egl.h:151
eglCopyBuffers = _lib.eglCopyBuffers
eglCopyBuffers.restype = EGLBoolean
eglCopyBuffers.argtypes = [EGLDisplay, EGLSurface, EGLNativePixmapType]

# /usr/include/EGL/egl.h:152
eglCreateContext = _lib.eglCreateContext
eglCreateContext.restype = EGLContext
eglCreateContext.argtypes = [EGLDisplay, EGLConfig, EGLContext, POINTER(EGLint)]

# /usr/include/EGL/egl.h:153
eglCreatePbufferSurface = _lib.eglCreatePbufferSurface
eglCreatePbufferSurface.restype = EGLSurface
eglCreatePbufferSurface.argtypes = [EGLDisplay, EGLConfig, POINTER(EGLint)]

# /usr/include/EGL/egl.h:154
eglCreatePixmapSurface = _lib.eglCreatePixmapSurface
eglCreatePixmapSurface.restype = EGLSurface
eglCreatePixmapSurface.argtypes = [EGLDisplay, EGLConfig, EGLNativePixmapType, POINTER(EGLint)]

# /usr/include/EGL/egl.h:155
eglCreateWindowSurface = _lib.eglCreateWindowSurface
eglCreateWindowSurface.restype = EGLSurface
eglCreateWindowSurface.argtypes = [EGLDisplay, EGLConfig, EGLNativeWindowType, POINTER(EGLint)]

# /usr/include/EGL/egl.h:156
eglDestroyContext = _lib.eglDestroyContext
eglDestroyContext.restype = EGLBoolean
eglDestroyContext.argtypes = [EGLDisplay, EGLContext]

# /usr/include/EGL/egl.h:157
eglDestroySurface = _lib.eglDestroySurface
eglDestroySurface.restype = EGLBoolean
eglDestroySurface.argtypes = [EGLDisplay, EGLSurface]

# /usr/include/EGL/egl.h:158
eglGetConfigAttrib = _lib.eglGetConfigAttrib
eglGetConfigAttrib.restype = EGLBoolean
eglGetConfigAttrib.argtypes = [EGLDisplay, EGLConfig, EGLint, POINTER(EGLint)]

# /usr/include/EGL/egl.h:159
eglGetConfigs = _lib.eglGetConfigs
eglGetConfigs.restype = EGLBoolean
eglGetConfigs.argtypes = [EGLDisplay, POINTER(EGLConfig), EGLint, POINTER(EGLint)]

# /usr/include/EGL/egl.h:160
eglGetCurrentDisplay = _lib.eglGetCurrentDisplay
eglGetCurrentDisplay.restype = EGLDisplay
eglGetCurrentDisplay.argtypes = []

# /usr/include/EGL/egl.h:161
eglGetCurrentSurface = _lib.eglGetCurrentSurface
eglGetCurrentSurface.restype = EGLSurface
eglGetCurrentSurface.argtypes = [EGLint]

# /usr/include/EGL/egl.h:162
eglGetDisplay = _lib.eglGetDisplay
eglGetDisplay.restype = EGLDisplay
eglGetDisplay.argtypes = [EGLNativeDisplayType]

# /usr/include/EGL/egl.h:163
eglGetError = _lib.eglGetError
eglGetError.restype = EGLint
eglGetError.argtypes = []

# /usr/include/EGL/egl.h:164
eglGetProcAddress = _lib.eglGetProcAddress
eglGetProcAddress.restype = __eglMustCastToProperFunctionPointerType
eglGetProcAddress.argtypes = [c_char_p]

# /usr/include/EGL/egl.h:165
eglInitialize = _lib.eglInitialize
eglInitialize.restype = EGLBoolean
eglInitialize.argtypes = [EGLDisplay, POINTER(EGLint), POINTER(EGLint)]

# /usr/include/EGL/egl.h:166
eglMakeCurrent = _lib.eglMakeCurrent
eglMakeCurrent.restype = EGLBoolean
eglMakeCurrent.argtypes = [EGLDisplay, EGLSurface, EGLSurface, EGLContext]

# /usr/include/EGL/egl.h:167
eglQueryContext = _lib.eglQueryContext
eglQueryContext.restype = EGLBoolean
eglQueryContext.argtypes = [EGLDisplay, EGLContext, EGLint, POINTER(EGLint)]

# /usr/include/EGL/egl.h:168
eglQueryString = _lib.eglQueryString
eglQueryString.restype = c_char_p
eglQueryString.argtypes = [EGLDisplay, EGLint]

# /usr/include/EGL/egl.h:169
eglQuerySurface = _lib.eglQuerySurface
eglQuerySurface.restype = EGLBoolean
eglQuerySurface.argtypes = [EGLDisplay, EGLSurface, EGLint, POINTER(EGLint)]

# /usr/include/EGL/egl.h:170
eglSwapBuffers = _lib.eglSwapBuffers
eglSwapBuffers.restype = EGLBoolean
eglSwapBuffers.argtypes = [EGLDisplay, EGLSurface]

# /usr/include/EGL/egl.h:171
eglTerminate = _lib.eglTerminate
eglTerminate.restype = EGLBoolean
eglTerminate.argtypes = [EGLDisplay]

# /usr/include/EGL/egl.h:172
eglWaitGL = _lib.eglWaitGL
eglWaitGL.restype = EGLBoolean
eglWaitGL.argtypes = []

# /usr/include/EGL/egl.h:173
eglWaitNative = _lib.eglWaitNative
eglWaitNative.restype = EGLBoolean
eglWaitNative.argtypes = [EGLint]

EGL_VERSION_1_1 = 1 	# /usr/include/EGL/egl.h:178
EGL_BACK_BUFFER = 12420 	# /usr/include/EGL/egl.h:179
EGL_BIND_TO_TEXTURE_RGB = 12345 	# /usr/include/EGL/egl.h:180
EGL_BIND_TO_TEXTURE_RGBA = 12346 	# /usr/include/EGL/egl.h:181
EGL_CONTEXT_LOST = 12302 	# /usr/include/EGL/egl.h:182
EGL_MIN_SWAP_INTERVAL = 12347 	# /usr/include/EGL/egl.h:183
EGL_MAX_SWAP_INTERVAL = 12348 	# /usr/include/EGL/egl.h:184
EGL_MIPMAP_TEXTURE = 12418 	# /usr/include/EGL/egl.h:185
EGL_MIPMAP_LEVEL = 12419 	# /usr/include/EGL/egl.h:186
EGL_NO_TEXTURE = 12380 	# /usr/include/EGL/egl.h:187
EGL_TEXTURE_2D = 12383 	# /usr/include/EGL/egl.h:188
EGL_TEXTURE_FORMAT = 12416 	# /usr/include/EGL/egl.h:189
EGL_TEXTURE_RGB = 12381 	# /usr/include/EGL/egl.h:190
EGL_TEXTURE_RGBA = 12382 	# /usr/include/EGL/egl.h:191
EGL_TEXTURE_TARGET = 12417 	# /usr/include/EGL/egl.h:192
PFNEGLBINDTEXIMAGEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLint) 	# /usr/include/EGL/egl.h:193
PFNEGLRELEASETEXIMAGEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLint) 	# /usr/include/EGL/egl.h:194
PFNEGLSURFACEATTRIBPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSurface, EGLint, EGLint) 	# /usr/include/EGL/egl.h:195
PFNEGLSWAPINTERVALPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLint) 	# /usr/include/EGL/egl.h:196
# /usr/include/EGL/egl.h:198
eglBindTexImage = _lib.eglBindTexImage
eglBindTexImage.restype = EGLBoolean
eglBindTexImage.argtypes = [EGLDisplay, EGLSurface, EGLint]

# /usr/include/EGL/egl.h:199
eglReleaseTexImage = _lib.eglReleaseTexImage
eglReleaseTexImage.restype = EGLBoolean
eglReleaseTexImage.argtypes = [EGLDisplay, EGLSurface, EGLint]

# /usr/include/EGL/egl.h:200
eglSurfaceAttrib = _lib.eglSurfaceAttrib
eglSurfaceAttrib.restype = EGLBoolean
eglSurfaceAttrib.argtypes = [EGLDisplay, EGLSurface, EGLint, EGLint]

# /usr/include/EGL/egl.h:201
eglSwapInterval = _lib.eglSwapInterval
eglSwapInterval.restype = EGLBoolean
eglSwapInterval.argtypes = [EGLDisplay, EGLint]

EGL_VERSION_1_2 = 1 	# /usr/include/EGL/egl.h:206
EGLenum = c_uint 	# /usr/include/EGL/egl.h:207
EGLClientBuffer = POINTER(None) 	# /usr/include/EGL/egl.h:208
EGL_ALPHA_FORMAT = 12424 	# /usr/include/EGL/egl.h:209
EGL_ALPHA_FORMAT_NONPRE = 12427 	# /usr/include/EGL/egl.h:210
EGL_ALPHA_FORMAT_PRE = 12428 	# /usr/include/EGL/egl.h:211
EGL_ALPHA_MASK_SIZE = 12350 	# /usr/include/EGL/egl.h:212
EGL_BUFFER_PRESERVED = 12436 	# /usr/include/EGL/egl.h:213
EGL_BUFFER_DESTROYED = 12437 	# /usr/include/EGL/egl.h:214
EGL_CLIENT_APIS = 12429 	# /usr/include/EGL/egl.h:215
EGL_COLORSPACE = 12423 	# /usr/include/EGL/egl.h:216
EGL_COLORSPACE_sRGB = 12425 	# /usr/include/EGL/egl.h:217
EGL_COLORSPACE_LINEAR = 12426 	# /usr/include/EGL/egl.h:218
EGL_COLOR_BUFFER_TYPE = 12351 	# /usr/include/EGL/egl.h:219
EGL_CONTEXT_CLIENT_TYPE = 12439 	# /usr/include/EGL/egl.h:220
EGL_DISPLAY_SCALING = 10000 	# /usr/include/EGL/egl.h:221
EGL_HORIZONTAL_RESOLUTION = 12432 	# /usr/include/EGL/egl.h:222
EGL_LUMINANCE_BUFFER = 12431 	# /usr/include/EGL/egl.h:223
EGL_LUMINANCE_SIZE = 12349 	# /usr/include/EGL/egl.h:224
EGL_OPENGL_ES_BIT = 1 	# /usr/include/EGL/egl.h:225
EGL_OPENVG_BIT = 2 	# /usr/include/EGL/egl.h:226
EGL_OPENGL_ES_API = 12448 	# /usr/include/EGL/egl.h:227
EGL_OPENVG_API = 12449 	# /usr/include/EGL/egl.h:228
EGL_OPENVG_IMAGE = 12438 	# /usr/include/EGL/egl.h:229
EGL_PIXEL_ASPECT_RATIO = 12434 	# /usr/include/EGL/egl.h:230
EGL_RENDERABLE_TYPE = 12352 	# /usr/include/EGL/egl.h:231
EGL_RENDER_BUFFER = 12422 	# /usr/include/EGL/egl.h:232
EGL_RGB_BUFFER = 12430 	# /usr/include/EGL/egl.h:233
EGL_SINGLE_BUFFER = 12421 	# /usr/include/EGL/egl.h:234
EGL_SWAP_BEHAVIOR = 12435 	# /usr/include/EGL/egl.h:235
EGL_VERTICAL_RESOLUTION = 12433 	# /usr/include/EGL/egl.h:237
PFNEGLBINDAPIPROC = CFUNCTYPE(EGLBoolean, EGLenum) 	# /usr/include/EGL/egl.h:238
PFNEGLQUERYAPIPROC = CFUNCTYPE(EGLenum) 	# /usr/include/EGL/egl.h:239
PFNEGLCREATEPBUFFERFROMCLIENTBUFFERPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLenum, EGLClientBuffer, EGLConfig, POINTER(EGLint)) 	# /usr/include/EGL/egl.h:240
PFNEGLRELEASETHREADPROC = CFUNCTYPE(EGLBoolean) 	# /usr/include/EGL/egl.h:241
PFNEGLWAITCLIENTPROC = CFUNCTYPE(EGLBoolean) 	# /usr/include/EGL/egl.h:242
# /usr/include/EGL/egl.h:244
eglBindAPI = _lib.eglBindAPI
eglBindAPI.restype = EGLBoolean
eglBindAPI.argtypes = [EGLenum]

# /usr/include/EGL/egl.h:245
eglQueryAPI = _lib.eglQueryAPI
eglQueryAPI.restype = EGLenum
eglQueryAPI.argtypes = []

# /usr/include/EGL/egl.h:246
eglCreatePbufferFromClientBuffer = _lib.eglCreatePbufferFromClientBuffer
eglCreatePbufferFromClientBuffer.restype = EGLSurface
eglCreatePbufferFromClientBuffer.argtypes = [EGLDisplay, EGLenum, EGLClientBuffer, EGLConfig, POINTER(EGLint)]

# /usr/include/EGL/egl.h:247
eglReleaseThread = _lib.eglReleaseThread
eglReleaseThread.restype = EGLBoolean
eglReleaseThread.argtypes = []

# /usr/include/EGL/egl.h:248
eglWaitClient = _lib.eglWaitClient
eglWaitClient.restype = EGLBoolean
eglWaitClient.argtypes = []

EGL_VERSION_1_3 = 1 	# /usr/include/EGL/egl.h:253
EGL_CONFORMANT = 12354 	# /usr/include/EGL/egl.h:254
EGL_CONTEXT_CLIENT_VERSION = 12440 	# /usr/include/EGL/egl.h:255
EGL_MATCH_NATIVE_PIXMAP = 12353 	# /usr/include/EGL/egl.h:256
EGL_OPENGL_ES2_BIT = 4 	# /usr/include/EGL/egl.h:257
EGL_VG_ALPHA_FORMAT = 12424 	# /usr/include/EGL/egl.h:258
EGL_VG_ALPHA_FORMAT_NONPRE = 12427 	# /usr/include/EGL/egl.h:259
EGL_VG_ALPHA_FORMAT_PRE = 12428 	# /usr/include/EGL/egl.h:260
EGL_VG_ALPHA_FORMAT_PRE_BIT = 64 	# /usr/include/EGL/egl.h:261
EGL_VG_COLORSPACE = 12423 	# /usr/include/EGL/egl.h:262
EGL_VG_COLORSPACE_sRGB = 12425 	# /usr/include/EGL/egl.h:263
EGL_VG_COLORSPACE_LINEAR = 12426 	# /usr/include/EGL/egl.h:264
EGL_VG_COLORSPACE_LINEAR_BIT = 32 	# /usr/include/EGL/egl.h:265
EGL_VERSION_1_4 = 1 	# /usr/include/EGL/egl.h:269
EGL_MULTISAMPLE_RESOLVE_BOX_BIT = 512 	# /usr/include/EGL/egl.h:271
EGL_MULTISAMPLE_RESOLVE = 12441 	# /usr/include/EGL/egl.h:272
EGL_MULTISAMPLE_RESOLVE_DEFAULT = 12442 	# /usr/include/EGL/egl.h:273
EGL_MULTISAMPLE_RESOLVE_BOX = 12443 	# /usr/include/EGL/egl.h:274
EGL_OPENGL_API = 12450 	# /usr/include/EGL/egl.h:275
EGL_OPENGL_BIT = 8 	# /usr/include/EGL/egl.h:276
EGL_SWAP_BEHAVIOR_PRESERVED_BIT = 1024 	# /usr/include/EGL/egl.h:277
PFNEGLGETCURRENTCONTEXTPROC = CFUNCTYPE(EGLContext) 	# /usr/include/EGL/egl.h:278
# /usr/include/EGL/egl.h:280
eglGetCurrentContext = _lib.eglGetCurrentContext
eglGetCurrentContext.restype = EGLContext
eglGetCurrentContext.argtypes = []

EGL_VERSION_1_5 = 1 	# /usr/include/EGL/egl.h:285
EGLSync = POINTER(None) 	# /usr/include/EGL/egl.h:286
intptr_t = c_long 	# /usr/include/stdint.h:87
EGLAttrib = intptr_t 	# /usr/include/EGL/egl.h:287
khronos_uint64_t = c_uint64 	# /usr/include/KHR/khrplatform.h:153
khronos_utime_nanoseconds_t = khronos_uint64_t 	# /usr/include/KHR/khrplatform.h:267
EGLTime = khronos_utime_nanoseconds_t 	# /usr/include/EGL/egl.h:288
EGLImage = POINTER(None) 	# /usr/include/EGL/egl.h:289
EGL_CONTEXT_MAJOR_VERSION = 12440 	# /usr/include/EGL/egl.h:290
EGL_CONTEXT_MINOR_VERSION = 12539 	# /usr/include/EGL/egl.h:291
EGL_CONTEXT_OPENGL_PROFILE_MASK = 12541 	# /usr/include/EGL/egl.h:292
EGL_CONTEXT_OPENGL_RESET_NOTIFICATION_STRATEGY = 12733 	# /usr/include/EGL/egl.h:293
EGL_NO_RESET_NOTIFICATION = 12734 	# /usr/include/EGL/egl.h:294
EGL_LOSE_CONTEXT_ON_RESET = 12735 	# /usr/include/EGL/egl.h:295
EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT = 1 	# /usr/include/EGL/egl.h:296
EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT = 2 	# /usr/include/EGL/egl.h:297
EGL_CONTEXT_OPENGL_DEBUG = 12720 	# /usr/include/EGL/egl.h:298
EGL_CONTEXT_OPENGL_FORWARD_COMPATIBLE = 12721 	# /usr/include/EGL/egl.h:299
EGL_CONTEXT_OPENGL_ROBUST_ACCESS = 12722 	# /usr/include/EGL/egl.h:300
EGL_OPENGL_ES3_BIT = 64 	# /usr/include/EGL/egl.h:301
EGL_CL_EVENT_HANDLE = 12444 	# /usr/include/EGL/egl.h:302
EGL_SYNC_CL_EVENT = 12542 	# /usr/include/EGL/egl.h:303
EGL_SYNC_CL_EVENT_COMPLETE = 12543 	# /usr/include/EGL/egl.h:304
EGL_SYNC_PRIOR_COMMANDS_COMPLETE = 12528 	# /usr/include/EGL/egl.h:305
EGL_SYNC_TYPE = 12535 	# /usr/include/EGL/egl.h:306
EGL_SYNC_STATUS = 12529 	# /usr/include/EGL/egl.h:307
EGL_SYNC_CONDITION = 12536 	# /usr/include/EGL/egl.h:308
EGL_SIGNALED = 12530 	# /usr/include/EGL/egl.h:309
EGL_UNSIGNALED = 12531 	# /usr/include/EGL/egl.h:310
EGL_SYNC_FLUSH_COMMANDS_BIT = 1 	# /usr/include/EGL/egl.h:311
EGL_FOREVER = 18446744073709551615 	# /usr/include/EGL/egl.h:312
EGL_TIMEOUT_EXPIRED = 12533 	# /usr/include/EGL/egl.h:313
EGL_CONDITION_SATISFIED = 12534 	# /usr/include/EGL/egl.h:314
EGL_SYNC_FENCE = 12537 	# /usr/include/EGL/egl.h:316
EGL_GL_COLORSPACE = 12445 	# /usr/include/EGL/egl.h:317
EGL_GL_COLORSPACE_SRGB = 12425 	# /usr/include/EGL/egl.h:318
EGL_GL_COLORSPACE_LINEAR = 12426 	# /usr/include/EGL/egl.h:319
EGL_GL_RENDERBUFFER = 12473 	# /usr/include/EGL/egl.h:320
EGL_GL_TEXTURE_2D = 12465 	# /usr/include/EGL/egl.h:321
EGL_GL_TEXTURE_LEVEL = 12476 	# /usr/include/EGL/egl.h:322
EGL_GL_TEXTURE_3D = 12466 	# /usr/include/EGL/egl.h:323
EGL_GL_TEXTURE_ZOFFSET = 12477 	# /usr/include/EGL/egl.h:324
EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_X = 12467 	# /usr/include/EGL/egl.h:325
EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_X = 12468 	# /usr/include/EGL/egl.h:326
EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_Y = 12469 	# /usr/include/EGL/egl.h:327
EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Y = 12470 	# /usr/include/EGL/egl.h:328
EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_Z = 12471 	# /usr/include/EGL/egl.h:329
EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Z = 12472 	# /usr/include/EGL/egl.h:330
EGL_IMAGE_PRESERVED = 12498 	# /usr/include/EGL/egl.h:331
PFNEGLCREATESYNCPROC = CFUNCTYPE(EGLSync, EGLDisplay, EGLenum, POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:333
PFNEGLDESTROYSYNCPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSync) 	# /usr/include/EGL/egl.h:334
PFNEGLCLIENTWAITSYNCPROC = CFUNCTYPE(EGLint, EGLDisplay, EGLSync, EGLint, EGLTime) 	# /usr/include/EGL/egl.h:335
PFNEGLGETSYNCATTRIBPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSync, EGLint, POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:336
PFNEGLCREATEIMAGEPROC = CFUNCTYPE(EGLImage, EGLDisplay, EGLContext, EGLenum, EGLClientBuffer, POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:337
PFNEGLDESTROYIMAGEPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLImage) 	# /usr/include/EGL/egl.h:338
PFNEGLGETPLATFORMDISPLAYPROC = CFUNCTYPE(EGLDisplay, EGLenum, POINTER(None), POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:339
PFNEGLCREATEPLATFORMWINDOWSURFACEPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLConfig, POINTER(None), POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:340
PFNEGLCREATEPLATFORMPIXMAPSURFACEPROC = CFUNCTYPE(EGLSurface, EGLDisplay, EGLConfig, POINTER(None), POINTER(EGLAttrib)) 	# /usr/include/EGL/egl.h:341
PFNEGLWAITSYNCPROC = CFUNCTYPE(EGLBoolean, EGLDisplay, EGLSync, EGLint) 	# /usr/include/EGL/egl.h:342
# /usr/include/EGL/egl.h:344
eglCreateSync = _lib.eglCreateSync
eglCreateSync.restype = EGLSync
eglCreateSync.argtypes = [EGLDisplay, EGLenum, POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:345
eglDestroySync = _lib.eglDestroySync
eglDestroySync.restype = EGLBoolean
eglDestroySync.argtypes = [EGLDisplay, EGLSync]

# /usr/include/EGL/egl.h:346
eglClientWaitSync = _lib.eglClientWaitSync
eglClientWaitSync.restype = EGLint
eglClientWaitSync.argtypes = [EGLDisplay, EGLSync, EGLint, EGLTime]

# /usr/include/EGL/egl.h:347
eglGetSyncAttrib = _lib.eglGetSyncAttrib
eglGetSyncAttrib.restype = EGLBoolean
eglGetSyncAttrib.argtypes = [EGLDisplay, EGLSync, EGLint, POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:348
eglCreateImage = _lib.eglCreateImage
eglCreateImage.restype = EGLImage
eglCreateImage.argtypes = [EGLDisplay, EGLContext, EGLenum, EGLClientBuffer, POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:349
eglDestroyImage = _lib.eglDestroyImage
eglDestroyImage.restype = EGLBoolean
eglDestroyImage.argtypes = [EGLDisplay, EGLImage]

# /usr/include/EGL/egl.h:350
eglGetPlatformDisplay = _lib.eglGetPlatformDisplay
eglGetPlatformDisplay.restype = EGLDisplay
eglGetPlatformDisplay.argtypes = [EGLenum, POINTER(None), POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:351
eglCreatePlatformWindowSurface = _lib.eglCreatePlatformWindowSurface
eglCreatePlatformWindowSurface.restype = EGLSurface
eglCreatePlatformWindowSurface.argtypes = [EGLDisplay, EGLConfig, POINTER(None), POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:352
eglCreatePlatformPixmapSurface = _lib.eglCreatePlatformPixmapSurface
eglCreatePlatformPixmapSurface.restype = EGLSurface
eglCreatePlatformPixmapSurface.argtypes = [EGLDisplay, EGLConfig, POINTER(None), POINTER(EGLAttrib)]

# /usr/include/EGL/egl.h:353
eglWaitSync = _lib.eglWaitSync
eglWaitSync.restype = EGLBoolean
eglWaitSync.argtypes = [EGLDisplay, EGLSync, EGLint]


__all__ = ['__egl_h_', 'EGL_EGL_PROTOTYPES', 'EGL_VERSION_1_0', 'EGLBoolean',
'EGLDisplay', 'EGLConfig', 'EGLSurface', 'EGLContext',
'__eglMustCastToProperFunctionPointerType', 'EGL_ALPHA_SIZE',
'EGL_BAD_ACCESS', 'EGL_BAD_ALLOC', 'EGL_BAD_ATTRIBUTE', 'EGL_BAD_CONFIG',
'EGL_BAD_CONTEXT', 'EGL_BAD_CURRENT_SURFACE', 'EGL_BAD_DISPLAY',
'EGL_BAD_MATCH', 'EGL_BAD_NATIVE_PIXMAP', 'EGL_BAD_NATIVE_WINDOW',
'EGL_BAD_PARAMETER', 'EGL_BAD_SURFACE', 'EGL_BLUE_SIZE', 'EGL_BUFFER_SIZE',
'EGL_CONFIG_CAVEAT', 'EGL_CONFIG_ID', 'EGL_CORE_NATIVE_ENGINE',
'EGL_DEPTH_SIZE', 'EGL_DRAW', 'EGL_EXTENSIONS', 'EGL_FALSE', 'EGL_GREEN_SIZE',
'EGL_HEIGHT', 'EGL_LARGEST_PBUFFER', 'EGL_LEVEL', 'EGL_MAX_PBUFFER_HEIGHT',
'EGL_MAX_PBUFFER_PIXELS', 'EGL_MAX_PBUFFER_WIDTH', 'EGL_NATIVE_RENDERABLE',
'EGL_NATIVE_VISUAL_ID', 'EGL_NATIVE_VISUAL_TYPE', 'EGL_NONE',
'EGL_NON_CONFORMANT_CONFIG', 'EGL_NOT_INITIALIZED', 'EGL_PBUFFER_BIT',
'EGL_PIXMAP_BIT', 'EGL_READ', 'EGL_RED_SIZE', 'EGL_SAMPLES',
'EGL_SAMPLE_BUFFERS', 'EGL_SLOW_CONFIG', 'EGL_STENCIL_SIZE', 'EGL_SUCCESS',
'EGL_SURFACE_TYPE', 'EGL_TRANSPARENT_BLUE_VALUE',
'EGL_TRANSPARENT_GREEN_VALUE', 'EGL_TRANSPARENT_RED_VALUE',
'EGL_TRANSPARENT_RGB', 'EGL_TRANSPARENT_TYPE', 'EGL_TRUE', 'EGL_VENDOR',
'EGL_VERSION', 'EGL_WIDTH', 'EGL_WINDOW_BIT', 'PFNEGLCHOOSECONFIGPROC',
'PFNEGLCOPYBUFFERSPROC', 'PFNEGLCREATECONTEXTPROC',
'PFNEGLCREATEPBUFFERSURFACEPROC', 'PFNEGLCREATEPIXMAPSURFACEPROC',
'PFNEGLCREATEWINDOWSURFACEPROC', 'PFNEGLDESTROYCONTEXTPROC',
'PFNEGLDESTROYSURFACEPROC', 'PFNEGLGETCONFIGATTRIBPROC',
'PFNEGLGETCONFIGSPROC', 'PFNEGLGETCURRENTDISPLAYPROC',
'PFNEGLGETCURRENTSURFACEPROC', 'PFNEGLGETDISPLAYPROC', 'PFNEGLGETERRORPROC',
'PFNEGLGETPROCADDRESSPROC', 'PFNEGLINITIALIZEPROC', 'PFNEGLMAKECURRENTPROC',
'PFNEGLQUERYCONTEXTPROC', 'PFNEGLQUERYSTRINGPROC', 'PFNEGLQUERYSURFACEPROC',
'PFNEGLSWAPBUFFERSPROC', 'PFNEGLTERMINATEPROC', 'PFNEGLWAITGLPROC',
'PFNEGLWAITNATIVEPROC', 'eglChooseConfig', 'eglCopyBuffers',
'eglCreateContext', 'eglCreatePbufferSurface', 'eglCreatePixmapSurface',
'eglCreateWindowSurface', 'eglDestroyContext', 'eglDestroySurface',
'eglGetConfigAttrib', 'eglGetConfigs', 'eglGetCurrentDisplay',
'eglGetCurrentSurface', 'eglGetDisplay', 'eglGetError', 'eglGetProcAddress',
'eglInitialize', 'eglMakeCurrent', 'eglQueryContext', 'eglQueryString',
'eglQuerySurface', 'eglSwapBuffers', 'eglTerminate', 'eglWaitGL',
'eglWaitNative', 'EGL_VERSION_1_1', 'EGL_BACK_BUFFER',
'EGL_BIND_TO_TEXTURE_RGB', 'EGL_BIND_TO_TEXTURE_RGBA', 'EGL_CONTEXT_LOST',
'EGL_MIN_SWAP_INTERVAL', 'EGL_MAX_SWAP_INTERVAL', 'EGL_MIPMAP_TEXTURE',
'EGL_MIPMAP_LEVEL', 'EGL_NO_TEXTURE', 'EGL_TEXTURE_2D', 'EGL_TEXTURE_FORMAT',
'EGL_TEXTURE_RGB', 'EGL_TEXTURE_RGBA', 'EGL_TEXTURE_TARGET',
'PFNEGLBINDTEXIMAGEPROC', 'PFNEGLRELEASETEXIMAGEPROC',
'PFNEGLSURFACEATTRIBPROC', 'PFNEGLSWAPINTERVALPROC', 'eglBindTexImage',
'eglReleaseTexImage', 'eglSurfaceAttrib', 'eglSwapInterval',
'EGL_VERSION_1_2', 'EGLenum', 'EGLClientBuffer', 'EGL_ALPHA_FORMAT',
'EGL_ALPHA_FORMAT_NONPRE', 'EGL_ALPHA_FORMAT_PRE', 'EGL_ALPHA_MASK_SIZE',
'EGL_BUFFER_PRESERVED', 'EGL_BUFFER_DESTROYED', 'EGL_CLIENT_APIS',
'EGL_COLORSPACE', 'EGL_COLORSPACE_sRGB', 'EGL_COLORSPACE_LINEAR',
'EGL_COLOR_BUFFER_TYPE', 'EGL_CONTEXT_CLIENT_TYPE', 'EGL_DISPLAY_SCALING',
'EGL_HORIZONTAL_RESOLUTION', 'EGL_LUMINANCE_BUFFER', 'EGL_LUMINANCE_SIZE',
'EGL_OPENGL_ES_BIT', 'EGL_OPENVG_BIT', 'EGL_OPENGL_ES_API', 'EGL_OPENVG_API',
'EGL_OPENVG_IMAGE', 'EGL_PIXEL_ASPECT_RATIO', 'EGL_RENDERABLE_TYPE',
'EGL_RENDER_BUFFER', 'EGL_RGB_BUFFER', 'EGL_SINGLE_BUFFER',
'EGL_SWAP_BEHAVIOR', 'EGL_VERTICAL_RESOLUTION', 'PFNEGLBINDAPIPROC',
'PFNEGLQUERYAPIPROC', 'PFNEGLCREATEPBUFFERFROMCLIENTBUFFERPROC',
'PFNEGLRELEASETHREADPROC', 'PFNEGLWAITCLIENTPROC', 'eglBindAPI',
'eglQueryAPI', 'eglCreatePbufferFromClientBuffer', 'eglReleaseThread',
'eglWaitClient', 'EGL_VERSION_1_3', 'EGL_CONFORMANT',
'EGL_CONTEXT_CLIENT_VERSION', 'EGL_MATCH_NATIVE_PIXMAP', 'EGL_OPENGL_ES2_BIT',
'EGL_VG_ALPHA_FORMAT', 'EGL_VG_ALPHA_FORMAT_NONPRE',
'EGL_VG_ALPHA_FORMAT_PRE', 'EGL_VG_ALPHA_FORMAT_PRE_BIT', 'EGL_VG_COLORSPACE',
'EGL_VG_COLORSPACE_sRGB', 'EGL_VG_COLORSPACE_LINEAR',
'EGL_VG_COLORSPACE_LINEAR_BIT', 'EGL_VERSION_1_4',
'EGL_MULTISAMPLE_RESOLVE_BOX_BIT', 'EGL_MULTISAMPLE_RESOLVE',
'EGL_MULTISAMPLE_RESOLVE_DEFAULT', 'EGL_MULTISAMPLE_RESOLVE_BOX',
'EGL_OPENGL_API', 'EGL_OPENGL_BIT', 'EGL_SWAP_BEHAVIOR_PRESERVED_BIT',
'PFNEGLGETCURRENTCONTEXTPROC', 'eglGetCurrentContext', 'EGL_VERSION_1_5',
'EGLSync', 'EGLAttrib', 'EGLTime', 'EGLImage', 'EGL_CONTEXT_MAJOR_VERSION',
'EGL_CONTEXT_MINOR_VERSION', 'EGL_CONTEXT_OPENGL_PROFILE_MASK',
'EGL_CONTEXT_OPENGL_RESET_NOTIFICATION_STRATEGY', 'EGL_NO_RESET_NOTIFICATION',
'EGL_LOSE_CONTEXT_ON_RESET', 'EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT',
'EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT', 'EGL_CONTEXT_OPENGL_DEBUG',
'EGL_CONTEXT_OPENGL_FORWARD_COMPATIBLE', 'EGL_CONTEXT_OPENGL_ROBUST_ACCESS',
'EGL_OPENGL_ES3_BIT', 'EGL_CL_EVENT_HANDLE', 'EGL_SYNC_CL_EVENT',
'EGL_SYNC_CL_EVENT_COMPLETE', 'EGL_SYNC_PRIOR_COMMANDS_COMPLETE',
'EGL_SYNC_TYPE', 'EGL_SYNC_STATUS', 'EGL_SYNC_CONDITION', 'EGL_SIGNALED',
'EGL_UNSIGNALED', 'EGL_SYNC_FLUSH_COMMANDS_BIT', 'EGL_FOREVER',
'EGL_TIMEOUT_EXPIRED', 'EGL_CONDITION_SATISFIED', 'EGL_SYNC_FENCE',
'EGL_GL_COLORSPACE', 'EGL_GL_COLORSPACE_SRGB', 'EGL_GL_COLORSPACE_LINEAR',
'EGL_GL_RENDERBUFFER', 'EGL_GL_TEXTURE_2D', 'EGL_GL_TEXTURE_LEVEL',
'EGL_GL_TEXTURE_3D', 'EGL_GL_TEXTURE_ZOFFSET',
'EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_X', 'EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_X',
'EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_Y', 'EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Y',
'EGL_GL_TEXTURE_CUBE_MAP_POSITIVE_Z', 'EGL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Z',
'EGL_IMAGE_PRESERVED', 'PFNEGLCREATESYNCPROC', 'PFNEGLDESTROYSYNCPROC',
'PFNEGLCLIENTWAITSYNCPROC', 'PFNEGLGETSYNCATTRIBPROC',
'PFNEGLCREATEIMAGEPROC', 'PFNEGLDESTROYIMAGEPROC',
'PFNEGLGETPLATFORMDISPLAYPROC', 'PFNEGLCREATEPLATFORMWINDOWSURFACEPROC',
'PFNEGLCREATEPLATFORMPIXMAPSURFACEPROC', 'PFNEGLWAITSYNCPROC',
'eglCreateSync', 'eglDestroySync', 'eglClientWaitSync', 'eglGetSyncAttrib',
'eglCreateImage', 'eglDestroyImage', 'eglGetPlatformDisplay',
'eglCreatePlatformWindowSurface', 'eglCreatePlatformPixmapSurface',
'eglWaitSync']



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

from ctypes import c_void_p, c_bool

from pyglet.libs.darwin.cocoapy import ObjCClass, ObjCSubclass, send_super
from pyglet.libs.darwin.cocoapy import NSUInteger, NSUIntegerEncoding
from pyglet.libs.darwin.cocoapy import NSRectEncoding


class PygletWindow_Implementation:
    PygletWindow = ObjCSubclass('NSWindow', 'PygletWindow')

    @PygletWindow.method('B')
    def canBecomeKeyWindow(self):
        return True

    # When the window is being resized, it enters into a mini event loop that
    # only looks at mouseDragged and mouseUp events, blocking everything else.
    # Among other things, this makes it impossible to run an NSTimer to call the
    # idle() function in order to update the view during the resize.  So we
    # override this method, called by the resizing event loop, and call the
    # idle() function from here.  This *almost* works.  I can't figure out what
    # is happening at the very beginning of a resize event.  The NSView's
    # viewWillStartLiveResize method is called and then nothing happens until
    # the mouse is dragged.  I think NSApplication's nextEventMatchingMask_etc
    # method is being called instead of this one.  I don't really feel like
    # subclassing NSApplication just to fix this.  Also, to prevent white flashes
    # while resizing, we must also call idle() from the view's reshape method.
    @PygletWindow.method(b'@'+NSUIntegerEncoding+b'@@B')
    def nextEventMatchingMask_untilDate_inMode_dequeue_(self, mask, date, mode, dequeue):
        if self.inLiveResize():
            # Call the idle() method while we're stuck in a live resize event.
            from pyglet import app
            if app.event_loop is not None:
                app.event_loop.idle()

        event = send_super(self, 'nextEventMatchingMask:untilDate:inMode:dequeue:',
                           mask, date, mode, dequeue,
                           superclass_name='NSWindow',
                           argtypes=[NSUInteger, c_void_p, c_void_p, c_bool])

        if event.value is None:
            return 0
        else:
            return event.value

    # Need this for set_size to not flash.
    @PygletWindow.method(b'd'+NSRectEncoding)
    def animationResizeTime_(self, newFrame):
        return 0.0


class PygletToolWindow_Implementation:
    PygletToolWindow = ObjCSubclass('NSPanel', 'PygletToolWindow')

    @PygletToolWindow.method(b'@'+NSUIntegerEncoding+b'@@B')
    def nextEventMatchingMask_untilDate_inMode_dequeue_(self, mask, date, mode, dequeue):
        if self.inLiveResize():
            # Call the idle() method while we're stuck in a live resize event.
            from pyglet import app
            if app.event_loop is not None:
                app.event_loop.idle()

        event = send_super(self, 'nextEventMatchingMask:untilDate:inMode:dequeue:',
                           mask, date, mode, dequeue, argtypes=[NSUInteger, c_void_p, c_void_p, c_bool])

        if event.value == None:
            return 0
        else:
            return event.value

    # Need this for set_size to not flash.
    @PygletToolWindow.method(b'd'+NSRectEncoding)
    def animationResizeTime_(self, newFrame):
        return 0.0


PygletWindow = ObjCClass('PygletWindow')
PygletToolWindow = ObjCClass('PygletToolWindow')



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

import ctypes

import pyglet
from pyglet.input import base
from pyglet.libs import win32
from pyglet.libs.win32 import dinput
from pyglet.libs.win32 import _kernel32

# These instance names are not defined anywhere, obtained by experiment.  The
# GUID names (which seem to be ideally what are needed) are wrong/missing for
# most of my devices.

_abs_instance_names = {
    0: 'x',
    1: 'y',
    2: 'z',
    3: 'rx', 
    4: 'ry',
    5: 'rz',
}

_rel_instance_names = {
    0: 'x',
    1: 'y',
    2: 'wheel',
}

_btn_instance_names = {}


def _create_control(object_instance):
    raw_name = object_instance.tszName
    type = object_instance.dwType
    instance = dinput.DIDFT_GETINSTANCE(type)

    if type & dinput.DIDFT_ABSAXIS:
        name = _abs_instance_names.get(instance)
        control = base.AbsoluteAxis(name, 0, 0xffff, raw_name)
    elif type & dinput.DIDFT_RELAXIS:
        name = _rel_instance_names.get(instance)
        control = base.RelativeAxis(name, raw_name)
    elif type & dinput.DIDFT_BUTTON:
        name = _btn_instance_names.get(instance)
        control = base.Button(name, raw_name)
    elif type & dinput.DIDFT_POV:
        control = base.AbsoluteAxis(base.AbsoluteAxis.HAT, 
                                    0, 0xffffffff, raw_name)
    else:
        return

    control._type = object_instance.dwType
    return control


class DirectInputDevice(base.Device):
    def __init__(self, display, device, device_instance):
        name = device_instance.tszInstanceName
        super(DirectInputDevice, self).__init__(display, name)

        self._type = device_instance.dwDevType & 0xff
        self._subtype = device_instance.dwDevType & 0xff00

        self._device = device
        self._init_controls()
        self._set_format()

    def _init_controls(self):
        self.controls = []
        self._device.EnumObjects(
            dinput.LPDIENUMDEVICEOBJECTSCALLBACK(self._object_enum), 
            None, dinput.DIDFT_ALL)

    def _object_enum(self, object_instance, arg):
        control = _create_control(object_instance.contents)
        if control:
            self.controls.append(control)        
        return dinput.DIENUM_CONTINUE

    def _set_format(self):
        if not self.controls:
            return

        object_formats = (dinput.DIOBJECTDATAFORMAT * len(self.controls))()
        offset = 0
        for object_format, control in zip(object_formats, self.controls):
            object_format.dwOfs = offset
            object_format.dwType = control._type
            offset += 4
             
        format = dinput.DIDATAFORMAT()
        format.dwSize = ctypes.sizeof(format)
        format.dwObjSize = ctypes.sizeof(dinput.DIOBJECTDATAFORMAT)
        format.dwFlags = 0
        format.dwDataSize = offset
        format.dwNumObjs = len(object_formats)
        format.rgodf = ctypes.cast(ctypes.pointer(object_formats),
                                   dinput.LPDIOBJECTDATAFORMAT)
        self._device.SetDataFormat(format)

        prop = dinput.DIPROPDWORD()
        prop.diph.dwSize = ctypes.sizeof(prop)
        prop.diph.dwHeaderSize = ctypes.sizeof(prop.diph)
        prop.diph.dwObj = 0
        prop.diph.dwHow = dinput.DIPH_DEVICE
        prop.dwData = 64 * ctypes.sizeof(dinput.DIDATAFORMAT)
        self._device.SetProperty(dinput.DIPROP_BUFFERSIZE, ctypes.byref(prop.diph))

    def open(self, window=None, exclusive=False):
        if not self.controls:
            return

        if window is None:
            # Pick any open window, or the shadow window if no windows
            # have been created yet.
            window = pyglet.gl._shadow_window
            for window in pyglet.app.windows:
                break

        flags = dinput.DISCL_BACKGROUND
        if exclusive:
            flags |= dinput.DISCL_EXCLUSIVE
        else:
            flags |= dinput.DISCL_NONEXCLUSIVE
        
        self._wait_object = _kernel32.CreateEventW(None, False, False, None)
        self._device.SetEventNotification(self._wait_object)
        pyglet.app.platform_event_loop.add_wait_object(self._wait_object, self._dispatch_events)

        self._device.SetCooperativeLevel(window._hwnd, flags)
        self._device.Acquire()

    def close(self):
        if not self.controls:
            return

        pyglet.app.platform_event_loop.remove_wait_object(self._wait_object)

        self._device.Unacquire()
        self._device.SetEventNotification(None)

        _kernel32.CloseHandle(self._wait_object)

    def get_controls(self):
        return self.controls

    def _dispatch_events(self):
        if not self.controls:
            return
        
        events = (dinput.DIDEVICEOBJECTDATA * 64)()
        n_events = win32.DWORD(len(events))
        try:
            self._device.GetDeviceData(ctypes.sizeof(dinput.DIDEVICEOBJECTDATA),
                                       ctypes.cast(ctypes.pointer(events),
                                                   dinput.LPDIDEVICEOBJECTDATA),
                                       ctypes.byref(n_events),
                                       0)
        except OSError:
            return

        for event in events[:n_events.value]:
            index = event.dwOfs // 4
            self.controls[index].value = event.dwData


_i_dinput = None


def _init_directinput():
    global _i_dinput
    if _i_dinput:
        return
    
    _i_dinput = dinput.IDirectInput8()
    module = _kernel32.GetModuleHandleW(None)
    dinput.DirectInput8Create(module, dinput.DIRECTINPUT_VERSION,
                              dinput.IID_IDirectInput8W, 
                              ctypes.byref(_i_dinput), None)


def get_devices(display=None):
    _init_directinput()
    _devices = []

    def _device_enum(device_instance, arg):
        device = dinput.IDirectInputDevice8()
        _i_dinput.CreateDevice(device_instance.contents.guidInstance,
                               ctypes.byref(device),
                               None)
        _devices.append(DirectInputDevice(display, 
                                          device, device_instance.contents))
        
        return dinput.DIENUM_CONTINUE

    _i_dinput.EnumDevices(dinput.DI8DEVCLASS_ALL, 
                          dinput.LPDIENUMDEVICESCALLBACK(_device_enum), 
                          None, dinput.DIEDFL_ATTACHEDONLY)
    return _devices


def _create_joystick(device):
    if device._type in (dinput.DI8DEVTYPE_JOYSTICK,
                        dinput.DI8DEVTYPE_1STPERSON,
                        dinput.DI8DEVTYPE_GAMEPAD):
        return base.Joystick(device)


def get_joysticks(display=None):
    return [joystick 
            for joystick 
            in [_create_joystick(device) 
                for device
                in get_devices(display)] 
            if joystick is not None]



import warnings
from pyglet.image import *
from pyglet.image.codecs import *
from pyglet import com
from pyglet.libs.win32.constants import *
from pyglet.libs.win32.types import *
from pyglet.libs.win32 import _kernel32 as kernel32
from pyglet.libs.win32 import _ole32 as ole32

CLSID_WICImagingFactory1 = com.GUID(0xcacaf262, 0x9370, 0x4615, 0xa1, 0x3b, 0x9f, 0x55, 0x39, 0xda, 0x4c, 0xa)
CLSID_WICImagingFactory2 = com.GUID(0x317d06e8, 0x5f24, 0x433d, 0xbd, 0xf7, 0x79, 0xce, 0x68, 0xd8, 0xab, 0xc2)

# This is available with Windows 7 with a Platform Update, but unable to detect as it wasn't a version change to the OS,
# but a KB update. Available in atleast 8+.
if WINDOWS_8_OR_GREATER:
    CLSID_WICImagingFactory = CLSID_WICImagingFactory2
else:
    CLSID_WICImagingFactory = CLSID_WICImagingFactory1

WICBitmapCreateCacheOption = UINT
WICBitmapNoCache = 0
WICBitmapCacheOnDemand = 0x1
WICBitmapCacheOnLoad = 0x2
WICBITMAPCREATECACHEOPTION_FORCE_DWORD = 0x7fffffff

WICBitmapPaletteType = UINT
WICBitmapPaletteTypeCustom = 0

WICBitmapTransformOptions = UINT
WICBitmapTransformRotate0 = 0
WICBitmapTransformRotate90 = 0x1
WICBitmapTransformRotate180 = 0x2
WICBitmapTransformRotate270 = 0x3
WICBitmapTransformFlipHorizontal = 0x8
WICBitmapTransformFlipVertical = 0x10

WICBitmapDitherType = UINT
WICBitmapDitherTypeNone = 0
WICBitmapDitherTypeSolid = 0
WICBitmapDitherTypeOrdered4x4 = 0x1
WICBitmapDitherTypeOrdered8x8 = 0x2
WICBitmapDitherTypeOrdered16x16 = 0x3
WICBitmapDitherTypeSpiral4x4 = 0x4
WICBitmapDitherTypeSpiral8x8 = 0x5
WICBitmapDitherTypeDualSpiral4x4 = 0x6
WICBitmapDitherTypeDualSpiral8x8 = 0x7
WICBitmapDitherTypeErrorDiffusion = 0x8
WICBITMAPDITHERTYPE_FORCE_DWORD = 0x7fffffff
WICBITMAPTRANSFORMOPTIONS_FORCE_DWORD = 0x7fffffff


WICDecodeOptions = UINT
WICDecodeMetadataCacheOnDemand = 0
WICDecodeMetadataCacheOnLoad = 0x1
WICMETADATACACHEOPTION_FORCE_DWORD = 0x7fffffff

# Different pixel formats.
REFWICPixelFormatGUID = com.GUID
GUID_WICPixelFormat1bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x01)
GUID_WICPixelFormat2bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x02)
GUID_WICPixelFormat4bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x03)
GUID_WICPixelFormat8bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x04)
GUID_WICPixelFormatBlackWhite = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x05)
GUID_WICPixelFormat2bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x06)
GUID_WICPixelFormat4bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x07)
GUID_WICPixelFormat8bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x08)
GUID_WICPixelFormat8bppAlpha = com.GUID(0xe6cd0116, 0xeeba, 0x4161, 0xaa, 0x85, 0x27, 0xdd, 0x9f, 0xb3, 0xa8, 0x95)
GUID_WICPixelFormat16bppBGR555 = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x09)
GUID_WICPixelFormat16bppBGR565 = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0a)
GUID_WICPixelFormat16bppBGRA5551 = com.GUID(0x05ec7c2b, 0xf1e6, 0x4961, 0xad, 0x46, 0xe1, 0xcc, 0x81, 0x0a, 0x87, 0xd2)
GUID_WICPixelFormat16bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0b)
GUID_WICPixelFormat24bppBGR = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0c)
GUID_WICPixelFormat24bppRGB = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0d)
GUID_WICPixelFormat32bppBGR = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0e)
GUID_WICPixelFormat32bppBGRA = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0f)
GUID_WICPixelFormat32bppPBGRA = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x10)
GUID_WICPixelFormat32bppRGB = com.GUID(0xd98c6b95, 0x3efe, 0x47d6, 0xbb, 0x25, 0xeb, 0x17, 0x48, 0xab, 0x0c, 0xf1)  # 7 platform update?
GUID_WICPixelFormat32bppRGBA = com.GUID(0xf5c7ad2d, 0x6a8d, 0x43dd, 0xa7, 0xa8, 0xa2, 0x99, 0x35, 0x26, 0x1a, 0xe9)
GUID_WICPixelFormat32bppPRGBA = com.GUID(0x3cc4a650, 0xa527, 0x4d37, 0xa9, 0x16, 0x31, 0x42, 0xc7, 0xeb, 0xed, 0xba)
GUID_WICPixelFormat48bppRGB = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x15)
GUID_WICPixelFormat48bppBGR = com.GUID(0xe605a384, 0xb468, 0x46ce, 0xbb, 0x2e, 0x36, 0xf1, 0x80, 0xe6, 0x43, 0x13)


class IWICComponentInfo(com.pIUnknown):
    _methods_ = [
        ('GetComponentType',
         com.STDMETHOD()),
        ('GetCLSID',
         com.STDMETHOD()),
        ('GetSigningStatus',
         com.STDMETHOD()),
        ('GetAuthor',
         com.STDMETHOD()),
        ('GetVendorGUID',
         com.STDMETHOD()),
        ('GetVersion',
         com.STDMETHOD()),
        ('GetSpecVersion',
         com.STDMETHOD()),
        ('GetFriendlyName',
         com.STDMETHOD())
    ]


class IWICPixelFormatInfo(IWICComponentInfo, com.pIUnknown):
    _methods_ = [
        ('GetFormatGUID',
         com.STDMETHOD(POINTER(com.GUID))),
        ('GetColorContext',
         com.STDMETHOD()),
        ('GetBitsPerPixel',
         com.STDMETHOD(POINTER(UINT))),
        ('GetChannelCount',
         com.STDMETHOD(POINTER(UINT))),
        ('GetChannelMask',
         com.STDMETHOD())
    ]


class IWICBitmapSource(com.pIUnknown):
    _methods_ = [
        ('GetSize',
         com.STDMETHOD(POINTER(UINT), POINTER(UINT))),
        ('GetPixelFormat',
         com.STDMETHOD(POINTER(REFWICPixelFormatGUID))),
        ('GetResolution',
         com.STDMETHOD(POINTER(DOUBLE), POINTER(DOUBLE))),
        ('CopyPalette',
         com.STDMETHOD()),
        ('CopyPixels',
         com.STDMETHOD(c_void_p, UINT, UINT, c_void_p)),
    ]


class IWICFormatConverter(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IWICBitmapSource, POINTER(REFWICPixelFormatGUID), WICBitmapDitherType, c_void_p, DOUBLE, WICBitmapPaletteType)),
        ('CanConvert',
         com.STDMETHOD(POINTER(REFWICPixelFormatGUID), POINTER(REFWICPixelFormatGUID), POINTER(BOOL))),
    ]


class IWICMetadataQueryReader(com.pIUnknown):
    _methods_ = [
        ('GetContainerFormat',
         com.STDMETHOD()),
        ('GetLocation',
         com.STDMETHOD()),
        ('GetMetadataByName',
         com.STDMETHOD(LPCWSTR, c_void_p)),
        ('GetEnumerator',
         com.STDMETHOD()),
    ]


class IWICBitmapFrameDecode(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('GetMetadataQueryReader',
         com.STDMETHOD(POINTER(IWICMetadataQueryReader))),
        ('GetColorContexts',
         com.STDMETHOD()),
        ('GetThumbnail',
         com.STDMETHOD(POINTER(IWICBitmapSource))),
    ]


class IWICBitmapFlipRotator(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IWICBitmapSource, WICBitmapTransformOptions)),
    ]


class IWICBitmap(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Lock',
         com.STDMETHOD()),
        ('SetPalette',
         com.STDMETHOD()),
        ('SetResolution',
         com.STDMETHOD())
    ]


class IWICBitmapDecoder(com.pIUnknown):
    _methods_ = [
        ('QueryCapability',
         com.STDMETHOD()),
        ('Initialize',
         com.STDMETHOD()),
        ('GetContainerFormat',
         com.STDMETHOD()),
        ('GetDecoderInfo',
         com.STDMETHOD()),
        ('CopyPalette',
         com.STDMETHOD()),
        ('GetMetadataQueryReader',
         com.STDMETHOD(POINTER(IWICMetadataQueryReader))),
        ('GetPreview',
         com.STDMETHOD()),
        ('GetColorContexts',
         com.STDMETHOD()),
        ('GetThumbnail',
         com.STDMETHOD()),
        ('GetFrameCount',
         com.STDMETHOD(POINTER(UINT))),
        ('GetFrame',
         com.STDMETHOD(UINT, POINTER(IWICBitmapFrameDecode))),
    ]


IID_IWICImagingFactory1 = com.GUID(0xec5ec8a9, 0xc395, 0x4314, 0x9c, 0x77, 0x54, 0xd7, 0xa9, 0x35, 0xff, 0x70)
IID_IWICImagingFactory2 = com.GUID(0x7B816B45, 0x1996, 0x4476, 0xB1, 0x32, 0xDE, 0x9E, 0x24, 0x7C, 0x8A, 0xF0)

if WINDOWS_8_OR_GREATER:
    IID_IWICImagingFactory = IID_IWICImagingFactory2
else:
    IID_IWICImagingFactory = IID_IWICImagingFactory1

IID_IWICPixelFormatInfo = com.GUID(0xE8EDA601, 0x3D48, 0x431a, 0xAB, 0x44, 0x69, 0x05, 0x9B, 0xE8, 0x8B, 0xBE)


class IWICImagingFactory(com.pIUnknown):
    _methods_ = [
        ('CreateDecoderFromFilename',
         com.STDMETHOD(LPCWSTR, com.GUID, DWORD, WICDecodeOptions, POINTER(IWICBitmapDecoder))),
        ('CreateDecoderFromStream',
         com.STDMETHOD(com.pIUnknown, c_void_p, WICDecodeOptions, POINTER(IWICBitmapDecoder))),
        ('CreateDecoderFromFileHandle',
         com.STDMETHOD()),
        ('CreateComponentInfo',
         com.STDMETHOD(com.GUID, POINTER(IWICComponentInfo))),
        ('CreateDecoder',
         com.STDMETHOD()),
        ('CreateEncoder',
         com.STDMETHOD()),
        ('CreatePalette',
         com.STDMETHOD()),
        ('CreateFormatConverter',
         com.STDMETHOD(POINTER(IWICFormatConverter))),
        ('CreateBitmapScaler',
         com.STDMETHOD()),
        ('CreateBitmapClipper',
         com.STDMETHOD()),
        ('CreateBitmapFlipRotator',
         com.STDMETHOD(POINTER(IWICBitmapFlipRotator))),
        ('CreateStream',
         com.STDMETHOD()),
        ('CreateColorContext',
         com.STDMETHOD()),
        ('CreateColorTransformer',
         com.STDMETHOD()),
        ('CreateBitmap',
         com.STDMETHOD(UINT, UINT, REFWICPixelFormatGUID, WICBitmapCreateCacheOption, POINTER(IWICBitmap))),
        ('CreateBitmapFromSource',
         com.STDMETHOD()),
        ('CreateBitmapFromSourceRect',
         com.STDMETHOD()),
        ('CreateBitmapFromMemory',
         com.STDMETHOD()),
        ('CreateBitmapFromHBITMAP',
         com.STDMETHOD()),
        ('CreateBitmapFromHICON',
         com.STDMETHOD()),
        ('CreateComponentEnumerator',
         com.STDMETHOD()),
        ('CreateFastMetadataEncoderFromDecoder',
         com.STDMETHOD()),
        ('CreateFastMetadataEncoderFromFrameDecode',
         com.STDMETHOD()),
        ('CreateQueryWriter',
         com.STDMETHOD()),
        ('CreateQueryWriterFromReader',
         com.STDMETHOD())
    ]


class WICDecoder(ImageDecoder):
    """Windows Imaging Component.
    This decoder is a replacement for GDI and GDI+ starting with Windows 7 with more features up to Windows 10."""
    def __init__(self):
        super(ImageDecoder, self).__init__()
        self._factory = IWICImagingFactory()

        try:
            ole32.CoInitializeEx(None, COINIT_MULTITHREADED)
        except OSError as err:
            warnings.warn(str(err))

        ole32.CoCreateInstance(CLSID_WICImagingFactory,
                               None,
                               CLSCTX_INPROC_SERVER,
                               IID_IWICImagingFactory,
                               byref(self._factory))

    def get_file_extensions(self):
        return ['.bmp', '.jpg', '.jpeg', '.png', '.tif', '.tiff', '.ico', '.jxr', '.hdp', '.wdp']

    def _load_bitmap_decoder(self, file, filename):
        data = file.read()

        # Create a HGLOBAL with image data
        hglob = kernel32.GlobalAlloc(GMEM_MOVEABLE, len(data))
        ptr = kernel32.GlobalLock(hglob)
        memmove(ptr, data, len(data))
        kernel32.GlobalUnlock(hglob)

        # Create IStream for the HGLOBAL
        stream = com.pIUnknown()
        ole32.CreateStreamOnHGlobal(hglob, True, byref(stream))

        # Load image from stream
        decoder = IWICBitmapDecoder()
        status = self._factory.CreateDecoderFromStream(stream, None, WICDecodeMetadataCacheOnDemand, byref(decoder))
        if status != 0:
            stream.Release()
            raise ImageDecodeException('WIC cannot load %r' % (filename or file))

        return decoder, stream

    def _get_bitmap_frame(self, bitmap_decoder, frame_index):
        bitmap = IWICBitmapFrameDecode()
        bitmap_decoder.GetFrame(frame_index, byref(bitmap))
        return bitmap

    def _get_image(self, bitmap, target_fmt=GUID_WICPixelFormat32bppBGRA):
        """Get's image from bitmap, specifying target format, bitmap is released before returning."""
        width = UINT()
        height = UINT()

        bitmap.GetSize(byref(width), byref(height))

        width = int(width.value)
        height = int(height.value)

        # Get image pixel format
        pf = com.GUID(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        bitmap.GetPixelFormat(byref(pf))

        fmt = 'BGRA'
        # If target format is not what we want (32bit BGRA) convert it.
        if pf != target_fmt:
            converter = IWICFormatConverter()
            self._factory.CreateFormatConverter(byref(converter))

            conversion_possible = BOOL()
            converter.CanConvert(pf, target_fmt, byref(conversion_possible))

            # 99% of the time conversion will be possible to default.
            # However, we check to be safe and fallback to 24 bit BGR if not possible.
            if not conversion_possible:
                target_fmt = GUID_WICPixelFormat24bppBGR
                fmt = 'BGR'

            converter.Initialize(bitmap, target_fmt, WICBitmapDitherTypeNone, None, 0, WICBitmapPaletteTypeCustom)

            bitmap.Release()
            bitmap = converter

        # Most images are loaded with a negative pitch, which requires list comprehension to fix.
        # Create a flipped bitmap through the decoder rather through Python to increase performance.
        flipper = IWICBitmapFlipRotator()
        self._factory.CreateBitmapFlipRotator(byref(flipper))

        flipper.Initialize(bitmap, WICBitmapTransformFlipVertical)

        stride = len(fmt) * width
        buffer_size = stride * height

        buffer = (BYTE * buffer_size)()

        flipper.CopyPixels(None, stride, buffer_size, byref(buffer))

        flipper.Release()
        bitmap.Release()  # Can be converter.

        return ImageData(width, height, fmt, buffer)

    def _delete_bitmap_decoder(self, bitmap_decoder, stream):
        # Release decoder and stream
        bitmap_decoder.Release()
        stream.Release()

    def decode(self, file, filename):
        bitmap_decoder, stream = self._load_bitmap_decoder(file, filename)
        bitmap = self._get_bitmap_frame(bitmap_decoder, 0)
        image = self._get_image(bitmap)
        self._delete_bitmap_decoder(bitmap_decoder, stream)
        return image

    @staticmethod
    def get_property_value(reader, metadata_name):
        """
            Uses a metadata name and reader to return a single value. Can be used to get metadata from images.
            If failure, returns 0.
            Also handles cleanup of PROPVARIANT.
        """
        try:
            prop = PROPVARIANT()
            reader.GetMetadataByName(metadata_name, byref(prop))
            value = prop.llVal
            ole32.PropVariantClear(byref(prop))
        except OSError:
            value = 0

        return value


def get_decoders():
    return [WICDecoder()]


def get_encoders():
    return []



import sys

import pyglet
import pyglet.app
import pyglet.canvas
import pyglet.extlibs
import pyglet.libs
import pyglet.font
import pyglet.gl
import pyglet.graphics
import pyglet.image
import pyglet.image.codecs
import pyglet.input
import pyglet.media
import pyglet.media.drivers
if sys.platform != 'darwin':  # FIXME: Upstream library loader doesn't consider CONDA_PREFIX
    import pyglet.media.codecs.ffmpeg
import pyglet.window
import pyglet.text
import pyglet.text.formats

# Skip to avoid having to install audio drivers.
# import pyglet.media.drivers.openal
# import pyglet.media.drivers.pulse

# Platform-specific tests.

if sys.platform == 'win32':
    import pyglet.libs.win32
    import pyglet.window.win32
    import pyglet.media.drivers.directsound

if sys.platform == 'darwin':
    import pyglet.libs.darwin
    import pyglet.libs.darwin.cocoapy
    import pyglet.window.cocoa

if sys.platform == 'linux':
    import pyglet.libs.x11
    import pyglet.window.xlib



from pyglet.media.events import MediaEvent

import pyglet
import ctypes
from pyglet.libs.win32.constants import *
from pyglet.libs.win32.types import *
from pyglet import com
import platform
import os
from pyglet.util import debug_print

_debug = debug_print('debug_media')


def load_xaudio2(dll_name):
    """This will attempt to load a version of XAudio2. Versions supported: 2.9, 2.8.
       While Windows 8 ships with 2.8 and Windows 10 ships with version 2.9, it is possible to install 2.9 on 8/8.1.
    """
    xaudio2 = dll_name
    # System32 and SysWOW64 folders are opposite perception in Windows x64.
    # System32 = x64 dll's | SysWOW64 = x86 dlls
    # By default ctypes only seems to look in system32 regardless of Python architecture, which has x64 dlls.
    if platform.architecture()[0] == '32bit':
        if platform.machine().endswith('64'):  # Machine is 64 bit, Python is 32 bit.
            xaudio2 = os.path.join(os.environ['WINDIR'], 'SysWOW64', '{}.dll'.format(xaudio2))

    xaudio2_lib = ctypes.windll.LoadLibrary(xaudio2)

    # Somehow x3d uses different calling structure than the rest of the DLL; Only affects 32 bit? Microsoft...
    x3d_lib = ctypes.cdll.LoadLibrary(xaudio2)
    return xaudio2_lib, x3d_lib


try:
    xaudio2_lib, x3d_lib = load_xaudio2("xaudio2_9")
except OSError:
    _debug("Could not load XAudio2.9 library")
    try:
        xaudio2_lib, x3d_lib = load_xaudio2("xaudio2_8")
    except OSError:
        _debug("Could not load XAudio2.8 library")
        raise ImportError('Could not locate a supported XAudio2 library.')


UINT32 = c_uint32
FLOAT32 = c_float


class XAUDIO2_DEBUG_CONFIGURATION(ctypes.Structure):
    _fields_ = [
        ('TraceMask', UINT32),
        ('BreakMask', UINT32),
        ('LogThreadID', BOOL),
        ('LogFileline', BOOL),
        ('LogFunctionName', BOOL),
        ('LogTiming', BOOL),
    ]


class XAUDIO2_PERFORMANCE_DATA(ctypes.Structure):
    _fields_ = [
        ('AudioCyclesSinceLastQuery', c_uint64),
        ('TotalCyclesSinceLastQuery', c_uint64),
        ('MinimumCyclesPerQuantum', UINT32),
        ('MaximumCyclesPerQuantum', UINT32),
        ('MemoryUsageInBytes', UINT32),
        ('CurrentLatencyInSamples', UINT32),
        ('GlitchesSinceEngineStarted', UINT32),
        ('ActiveSourceVoiceCount', UINT32),
        ('TotalSourceVoiceCount', UINT32),
        ('ActiveSubmixVoiceCount', UINT32),
        ('ActiveResamplerCount', UINT32),
        ('ActiveMatrixMixCount', UINT32),
        ('ActiveXmaSourceVoices', UINT32),
        ('ActiveXmaStreams', UINT32),
    ]

    def __repr__(self):
        return "XAUDIO2PerformanceData(active_voices={}, total_voices={}, glitches={}, latency={} samples, memory_usage={} bytes)".format(self.ActiveSourceVoiceCount, self.TotalSourceVoiceCount, self.GlitchesSinceEngineStarted, self.CurrentLatencyInSamples, self.MemoryUsageInBytes)


class XAUDIO2_VOICE_SENDS(ctypes.Structure):
    _fields_ = [
        ('SendCount', UINT32),
        ('pSends', c_void_p),
    ]


class XAUDIO2_BUFFER(ctypes.Structure):
    _fields_ = [
        ('Flags', UINT32),
        ('AudioBytes', UINT32),
        ('pAudioData', POINTER(c_char)),
        ('PlayBegin', UINT32),
        ('PlayLength', UINT32),
        ('LoopBegin', UINT32),
        ('LoopLength', UINT32),
        ('LoopCount', UINT32),
        ('pContext', c_void_p),
    ]

class XAUDIO2_VOICE_STATE(ctypes.Structure):
    _fields_ = [
        ('pCurrentBufferContext', c_void_p),
        ('BuffersQueued', UINT32),
        ('SamplesPlayed', UINT32)
    ]

    def __repr__(self):
        return "XAUDIO2_VOICE_STATE(BuffersQueued={0}, SamplesPlayed={1})".format(self.BuffersQueued, self.SamplesPlayed)

class WAVEFORMATEX(ctypes.Structure):
    _fields_ = [
        ('wFormatTag', WORD),
        ('nChannels', WORD),
        ('nSamplesPerSec', DWORD),
        ('nAvgBytesPerSec', DWORD),
        ('nBlockAlign', WORD),
        ('wBitsPerSample', WORD),
        ('cbSize', WORD),
    ]

    def __repr__(self):
        return 'WAVEFORMATEX(wFormatTag={}, nChannels={}, nSamplesPerSec={}, nAvgBytesPersec={}' \
               ', nBlockAlign={}, wBitsPerSample={}, cbSize={})'.format(
            self.wFormatTag, self.nChannels, self.nSamplesPerSec,
            self.nAvgBytesPerSec, self.nBlockAlign, self.wBitsPerSample,
            self.cbSize)

XAUDIO2_USE_DEFAULT_PROCESSOR = 0x00000000  # Win 10+

if WINDOWS_10_ANNIVERSARY_UPDATE_OR_GREATER:
    XAUDIO2_DEFAULT_PROCESSOR = XAUDIO2_USE_DEFAULT_PROCESSOR
else:
    XAUDIO2_DEFAULT_PROCESSOR = 0x00000001  # Windows 8/8.1


XAUDIO2_LOG_ERRORS = 0x0001   # For handled errors with serious effects.
XAUDIO2_LOG_WARNINGS = 0x0002   # For handled errors that may be recoverable.
XAUDIO2_LOG_INFO = 0x0004   # Informational chit-chat (e.g. state changes).
XAUDIO2_LOG_DETAIL = 0x0008   # More detailed chit-chat.
XAUDIO2_LOG_API_CALLS = 0x0010   # Public API function entries and exits.
XAUDIO2_LOG_FUNC_CALLS = 0x0020   # Internal function entries and exits.
XAUDIO2_LOG_TIMING = 0x0040   # Delays detected and other timing data.
XAUDIO2_LOG_LOCKS = 0x0080   # Usage of critical sections and mutexes.
XAUDIO2_LOG_MEMORY = 0x0100   # Memory heap usage information.
XAUDIO2_LOG_STREAMING = 0x1000   # Audio streaming information.


# Some XAUDIO2 global settings, most not used, but useful information
XAUDIO2_MAX_BUFFER_BYTES = 0x80000000   # Maximum bytes allowed in a source buffer
XAUDIO2_MAX_QUEUED_BUFFERS = 64         # Maximum buffers allowed in a voice queue
XAUDIO2_MAX_BUFFERS_SYSTEM = 2          # Maximum buffers allowed for system threads (Xbox 360 only)
XAUDIO2_MAX_AUDIO_CHANNELS = 64         # Maximum channels in an audio stream
XAUDIO2_MIN_SAMPLE_RATE = 1000          # Minimum audio sample rate supported
XAUDIO2_MAX_SAMPLE_RATE = 200000        # Maximum audio sample rate supported
XAUDIO2_MAX_VOLUME_LEVEL = 16777216.0   # Maximum acceptable volume level (2^24)
XAUDIO2_MIN_FREQ_RATIO =  (1/1024.0)    # Minimum SetFrequencyRatio argument
XAUDIO2_MAX_FREQ_RATIO = 1024.0         # Maximum MaxFrequencyRatio argument
XAUDIO2_DEFAULT_FREQ_RATIO  = 2.0       # Default MaxFrequencyRatio argument
XAUDIO2_MAX_FILTER_ONEOVERQ = 1.5       # Maximum XAUDIO2_FILTER_PARAMETERS.OneOverQ
XAUDIO2_MAX_FILTER_FREQUENCY = 1.0      # Maximum XAUDIO2_FILTER_PARAMETERS.Frequency
XAUDIO2_MAX_LOOP_COUNT = 254            # Maximum non-infinite XAUDIO2_BUFFER.LoopCount
XAUDIO2_MAX_INSTANCES = 8               # Maximum simultaneous XAudio2 objects on Xbox 360


XAUDIO2_FILTER_TYPE = UINT
LowPassFilter = 0  # Attenuates frequencies above the cutoff frequency (state-variable filter).
BandPassFilter = 1   # Attenuates frequencies outside a given range      (state-variable filter).
HighPassFilter = 2   # Attenuates frequencies below the cutoff frequency (state-variable filter).
NotchFilter = 3   # Attenuates frequencies inside a given range       (state-variable filter).
LowPassOnePoleFilter = 4   # Attenuates frequencies above the cutoff frequency (one-pole filter, XAUDIO2_FILTER_PARAMETERS.OneOverQ has no effect)
HighPassOnePoleFilter = 5   # Attenuates frequencies below the cutoff frequency (one-pole filter, XAUDIO2_FILTER_PARAMETERS.OneOverQ has no effect)

XAUDIO2_NO_LOOP_REGION = 0  # Used in XAUDIO2_BUFFER.LoopCount
XAUDIO2_LOOP_INFINITE = 255  # Used in XAUDIO2_BUFFER.LoopCount
XAUDIO2_DEFAULT_CHANNELS = 0  # Used in CreateMasteringVoice
XAUDIO2_DEFAULT_SAMPLERATE = 0  # Used in CreateMasteringVoice

WAVE_FORMAT_PCM = 1

XAUDIO2_DEBUG_ENGINE =            0x0001    # Used in XAudio2Create
XAUDIO2_VOICE_NOPITCH =           0x0002    # Used in IXAudio2::CreateSourceVoice
XAUDIO2_VOICE_NOSRC =             0x0004    # Used in IXAudio2::CreateSourceVoice
XAUDIO2_VOICE_USEFILTER =         0x0008    # Used in IXAudio2::CreateSource/SubmixVoice
XAUDIO2_PLAY_TAILS =              0x0020    # Used in IXAudio2SourceVoice::Stop
XAUDIO2_END_OF_STREAM =           0x0040    # Used in XAUDIO2_BUFFER.Flags
XAUDIO2_SEND_USEFILTER =          0x0080    # Used in XAUDIO2_SEND_DESCRIPTOR.Flags
XAUDIO2_VOICE_NOSAMPLESPLAYED =   0x0100    # Used in IXAudio2SourceVoice::GetState
XAUDIO2_STOP_ENGINE_WHEN_IDLE  =  0x2000    # Used in XAudio2Create to force the engine to Stop when no source voices are Started, and Start when a voice is Started
XAUDIO2_1024_QUANTUM =            0x8000    # Used in XAudio2Create to specify nondefault processing quantum of 21.33 ms (1024 samples at 48KHz)
XAUDIO2_NO_VIRTUAL_AUDIO_CLIENT = 0x10000   # Used in CreateMasteringVoice to create a virtual audio client


class IXAudio2VoiceCallback(com.Interface):
    _methods_ = [
        ('OnVoiceProcessingPassStart',
         com.STDMETHOD(UINT32)),
        ('OnVoiceProcessingPassEnd',
         com.STDMETHOD()),
        ('onStreamEnd',
         com.STDMETHOD()),
        ('onBufferStart',
         com.STDMETHOD(ctypes.c_void_p)),
        ('OnBufferEnd',
         com.STDMETHOD(ctypes.c_void_p)),
        ('OnLoopEnd',
         com.STDMETHOD(ctypes.c_void_p)),
    ]


class XA2SourceCallback(com.COMObject):
    """Callback class used to trigger when buffers or streams end..
           WARNING: Whenever a callback is running, XAudio2 cannot generate audio.
           Make sure these functions run as fast as possible and do not block/delay more than a few milliseconds.
           MS Recommendation:
           At a minimum, callback functions must not do the following:
                - Access the hard disk or other permanent storage
                - Make expensive or blocking API calls
                - Synchronize with other parts of client code
                - Require significant CPU usage
    """
    _interfaces_ = [IXAudio2VoiceCallback]

    def __init__(self, xa2_player):
        self.xa2_player = xa2_player

    def OnVoiceProcessingPassStart(self, bytesRequired):
        pass

    def OnVoiceProcessingPassEnd(self):
        pass

    def onStreamEnd(self):
        pass

    def onBufferStart(self, pBufferContext):
        pass

    def OnBufferEnd(self, pBufferContext):
        """At the end of playing one buffer, attempt to refill again.
        Even if the player is out of sources, it needs to be called to purge all buffers.
        """
        if self.xa2_player:
            self.xa2_player.refill_source_player()

    def OnLoopEnd(self, this, pBufferContext):
        pass

    def onVoiceError(self, this, pBufferContext, hresult):
        raise Exception("Error occurred during audio playback.", hresult)


class XAUDIO2_EFFECT_DESCRIPTOR(Structure):
    _fields_ = [
        ('pEffect', com.pIUnknown),
        ('InitialState', c_bool),
        ('OutputChannels', UINT32)
    ]


class XAUDIO2_EFFECT_CHAIN(ctypes.Structure):
    _fields_ = [
        ('EffectCount', UINT32),
        ('pEffectDescriptors', POINTER(XAUDIO2_EFFECT_DESCRIPTOR)),
    ]


class XAUDIO2_FILTER_PARAMETERS(Structure):
    _fields_ = [
        ('Type', XAUDIO2_FILTER_TYPE),
        ('Frequency', FLOAT),
        ('OneOverQ', FLOAT)
    ]


class XAUDIO2_VOICE_DETAILS(Structure):
    _fields_ = [
        ('CreationFlags', UINT32),
        ('ActiveFlags', UINT32),
        ('InputChannels', UINT32),
        ('InputSampleRate', UINT32)
    ]


class IXAudio2Voice(com.pInterface):
    _methods_ = [
        ('GetVoiceDetails',
         com.STDMETHOD(POINTER(XAUDIO2_VOICE_DETAILS))),
        ('SetOutputVoices',
         com.STDMETHOD()),
        ('SetEffectChain',
         com.STDMETHOD(POINTER(XAUDIO2_EFFECT_CHAIN))),
        ('EnableEffect',
         com.STDMETHOD()),
        ('DisableEffect',
         com.STDMETHOD()),
        ('GetEffectState',
         com.STDMETHOD()),
        ('SetEffectParameters',
         com.STDMETHOD()),
        ('GetEffectParameters',
         com.STDMETHOD()),
        ('SetFilterParameters',
         com.STDMETHOD(POINTER(XAUDIO2_FILTER_PARAMETERS), UINT32)),
        ('GetFilterParameters',
         com.STDMETHOD()),
        ('SetOutputFilterParameters',
         com.STDMETHOD()),
        ('GetOutputFilterParameters',
         com.STDMETHOD()),
        ('SetVolume',
         com.STDMETHOD(ctypes.c_float, UINT32)),
        ('GetVolume',
         com.STDMETHOD(POINTER(c_float))),
        ('SetChannelVolumes',
         com.STDMETHOD()),
        ('GetChannelVolumes',
         com.STDMETHOD()),
        ('SetOutputMatrix',
         com.STDMETHOD(c_void_p, UINT32, UINT32, POINTER(FLOAT), UINT32)),
        ('GetOutputMatrix',
         com.STDMETHOD()),
        ('DestroyVoice',
         com.STDMETHOD())
    ]


class IXAudio2SubmixVoice(IXAudio2Voice):
    pass


class IXAudio2SourceVoice(IXAudio2Voice):
    _methods_ = [
        ('Start',
         com.STDMETHOD(UINT32, UINT32)),
        ('Stop',
         com.STDMETHOD(UINT32, UINT32)),
        ('SubmitSourceBuffer',
         com.STDMETHOD(POINTER(XAUDIO2_BUFFER), c_void_p)),
        ('FlushSourceBuffers',
         com.STDMETHOD()),
        ('Discontinuity',
         com.STDMETHOD()),
        ('ExitLoop',
         com.STDMETHOD()),
        ('GetState',
         com.STDMETHOD(POINTER(XAUDIO2_VOICE_STATE), UINT32)),
        ('SetFrequencyRatio',
         com.STDMETHOD(FLOAT, UINT32)),
        ('GetFrequencyRatio',
         com.STDMETHOD(POINTER(c_float))),
        ('SetSourceSampleRate',
         com.STDMETHOD()),
    ]


class IXAudio2MasteringVoice(IXAudio2Voice):
    _methods_ = [
        ('GetChannelMask',
         com.STDMETHOD(POINTER(DWORD)))
    ]


class IXAudio2EngineCallback(com.Interface):
    _methods_ = [
        ('OnProcessingPassStart',
         com.METHOD(ctypes.c_void_p)),
        ('OnProcessingPassEnd',
         com.METHOD(ctypes.c_void_p)),
        ('OnCriticalError',
         com.METHOD(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_ulong)),
    ]


class XA2EngineCallback(com.COMObject):
    _interfaces_ = [IXAudio2EngineCallback]

    def OnProcessingPassStart(self):
        pass

    def OnProcessingPassEnd(self):
        pass

    def OnCriticalError(self, this, hresult):
        raise Exception("Critical Error:", hresult)



# -------------- 3D Audio Positioning----------
class X3DAUDIO_DISTANCE_CURVE_POINT(ctypes.Structure):
    _fields_ = [
        ('Distance', FLOAT32),
        ('DSPSetting', FLOAT32)
    ]


class X3DAUDIO_DISTANCE_CURVE(ctypes.Structure):
    _fields_ = [
        ('pPoints', POINTER(X3DAUDIO_DISTANCE_CURVE_POINT)),
        ('PointCount', UINT32)
    ]


class X3DAUDIO_VECTOR(ctypes.Structure):
    _fields_ = [
        ('x', c_float),
        ('y', c_float),
        ('z', c_float),
    ]



"""Cone:
   Specifies directionality for a listener or single-channel emitter by
   modifying DSP behaviour with respect to its front orientation.
   This is modeled using two sound cones: an inner cone and an outer cone.
   On/within the inner cone, DSP settings are scaled by the inner values.
   On/beyond the outer cone, DSP settings are scaled by the outer values.
   If on both the cones, DSP settings are scaled by the inner values only.
   Between the two cones, the scaler is linearly interpolated between the
   inner and outer values.  Set both cone angles to 0 or X3DAUDIO_2PI for
   omnidirectionality using only the outer or inner values respectively."""
class X3DAUDIO_CONE(Structure):
    _fields_ = [
        ('InnerAngle', FLOAT32),  # inner cone angle in radians, must be within [0.0f, X3DAUDIO_2PI]
        ('OuterAngle', FLOAT32),  # outer cone angle in radians, must be within [InnerAngle, X3DAUDIO_2PI]
        ('InnerVolume', FLOAT32),  # volume level scaler on/within inner cone, used only for matrix calculations, must be within [0.0f, 2.0f] when used
        ('OuterVolume', FLOAT32),  #  volume level scaler on/beyond outer cone, used only for matrix calculations, must be within [0.0f, 2.0f] when used
        ('InnerLPF', FLOAT32),  # LPF (both direct and reverb paths) coefficient subtrahend on/within inner cone, used only for LPF (both direct and reverb paths) calculations, must be within [0.0f, 1.0f] when used
        ('OuterLPF', FLOAT32),  # LPF (both direct and reverb paths) coefficient subtrahend on/beyond outer cone, used only for LPF (both direct and reverb paths) calculations, must be within [0.0f, 1.0f] when used
        ('InnerReverb', FLOAT32),  # reverb send level scaler on/within inner cone, used only for reverb calculations, must be within [0.0f, 2.0f] when used
        ('OuterReverb', FLOAT32)  # reverb send level scaler on/beyond outer cone, used only for reverb calculations, must be within [0.0f, 2.0f] when used
    ]


class X3DAUDIO_LISTENER(Structure):
    _fields_ = [
        ('OrientFront', X3DAUDIO_VECTOR),  # orientation of front direction, used only for matrix and delay calculations or listeners with cones for matrix, LPF (both direct and reverb paths), and reverb calculations, must be normalized when used
        ('OrientTop', X3DAUDIO_VECTOR),  # orientation of top direction, used only for matrix and delay calculations, must be orthonormal with OrientFront when used
        ('Position', X3DAUDIO_VECTOR),  # position in user-defined world units, does not affect Velocity
        ('Velocity', X3DAUDIO_VECTOR),  # velocity vector in user-defined world units/second, used only for doppler calculations, does not affect Position
        ('pCone', POINTER(X3DAUDIO_CONE))  # sound cone, used only for matrix, LPF (both direct and reverb paths), and reverb calculations, NULL specifies omnidirectionality
    ]


class X3DAUDIO_EMITTER(Structure):
    _fields_ = [
        ('pCone', POINTER(X3DAUDIO_CONE)),
        ('OrientFront', X3DAUDIO_VECTOR),
        ('OrientTop', X3DAUDIO_VECTOR),
        ('Position', X3DAUDIO_VECTOR),
        ('Velocity', X3DAUDIO_VECTOR),
        ('InnerRadius', FLOAT32),
        ('InnerRadiusAngle', FLOAT32),
        ('ChannelCount', UINT32),
        ('ChannelRadius', FLOAT32),
        ('pChannelAzimuths', POINTER(FLOAT32)),
        ('pVolumeCurve', POINTER(X3DAUDIO_DISTANCE_CURVE)),
        ('pLFECurve', POINTER(X3DAUDIO_DISTANCE_CURVE)),
        ('pLPFDirectCurve', POINTER(X3DAUDIO_DISTANCE_CURVE)),
        ('pLPFReverbCurve', POINTER(X3DAUDIO_DISTANCE_CURVE)),
        ('pReverbCurve', POINTER(X3DAUDIO_DISTANCE_CURVE)),
        ('CurveDistanceScaler', FLOAT32),
        ('DopplerScaler', FLOAT32)
    ]


class X3DAUDIO_DSP_SETTINGS(Structure):
    _fields_ = [
        ('pMatrixCoefficients', POINTER(FLOAT)),  # float array
        ('pDelayTimes', POINTER(FLOAT32)),
        ('SrcChannelCount', UINT32),
        ('DstChannelCount', UINT32),
        ('LPFDirectCoefficient', FLOAT32),
        ('LPFReverbCoefficient', FLOAT32),
        ('ReverbLevel', FLOAT32),
        ('DopplerFactor', FLOAT32),
        ('EmitterToListenerAngle', FLOAT32),
        ('EmitterToListenerDistance', FLOAT32),
        ('EmitterVelocityComponent', FLOAT32),
        ('ListenerVelocityComponent', FLOAT32)
    ]

# Other constants that may or may not be used in X3D.

SPEAKER_FRONT_LEFT             = 0x00000001
SPEAKER_FRONT_RIGHT            = 0x00000002
SPEAKER_FRONT_CENTER           = 0x00000004
SPEAKER_LOW_FREQUENCY          = 0x00000008
SPEAKER_BACK_LEFT              = 0x00000010
SPEAKER_BACK_RIGHT             = 0x00000020
SPEAKER_FRONT_LEFT_OF_CENTER   = 0x00000040
SPEAKER_FRONT_RIGHT_OF_CENTER  = 0x00000080
SPEAKER_BACK_CENTER            = 0x00000100
SPEAKER_SIDE_LEFT              = 0x00000200
SPEAKER_SIDE_RIGHT             = 0x00000400
SPEAKER_TOP_CENTER             = 0x00000800
SPEAKER_TOP_FRONT_LEFT         = 0x00001000
SPEAKER_TOP_FRONT_CENTER       = 0x00002000
SPEAKER_TOP_FRONT_RIGHT        = 0x00004000
SPEAKER_TOP_BACK_LEFT          = 0x00008000
SPEAKER_TOP_BACK_CENTER        = 0x00010000
SPEAKER_TOP_BACK_RIGHT         = 0x00020000
SPEAKER_RESERVED               = 0x7FFC0000  # bit mask locations reserved for future use
SPEAKER_ALL                    = 0x80000000  # used to specify that any possible permutation of speaker configurations

SPEAKER_MONO = SPEAKER_FRONT_CENTER
SPEAKER_STEREO = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT)
SPEAKER_2POINT1 = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_LOW_FREQUENCY)
SPEAKER_SURROUND = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_FRONT_CENTER | SPEAKER_BACK_CENTER)
SPEAKER_QUAD = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT)
SPEAKER_4POINT1 = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_LOW_FREQUENCY | SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT)
SPEAKER_5POINT1 = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_FRONT_CENTER | SPEAKER_LOW_FREQUENCY | SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT)
SPEAKER_7POINT1 = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_FRONT_CENTER | SPEAKER_LOW_FREQUENCY | SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT | SPEAKER_FRONT_LEFT_OF_CENTER | SPEAKER_FRONT_RIGHT_OF_CENTER)
SPEAKER_5POINT1_SURROUND = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_FRONT_CENTER | SPEAKER_LOW_FREQUENCY | SPEAKER_SIDE_LEFT  | SPEAKER_SIDE_RIGHT)
SPEAKER_7POINT1_SURROUND = (SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT | SPEAKER_FRONT_CENTER | SPEAKER_LOW_FREQUENCY | SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT | SPEAKER_SIDE_LEFT  | SPEAKER_SIDE_RIGHT)


DBL_DECIMAL_DIG = 17                      # # of decimal digits of rounding precision
DBL_DIG = 15                      # # of decimal digits of precision
DBL_EPSILON = 2.2204460492503131e-016  # smallest such that 1.0+DBL_EPSILON != 1.0
DBL_HAS_SUBNORM = 1                       # type does support subnormal numbers
DBL_MANT_DIG  = 53                      # # of bits in mantissa
DBL_MAX  = 1.7976931348623158e+308  # max value
DBL_MAX_10_EXP = 308                     # max decimal exponent
DBL_MAX_EXP = 1024                    # max binary exponent
DBL_MIN = 2.2250738585072014e-308  # min positive value
DBL_MIN_10_EXP = (-307)                  # min decimal exponent
DBL_MIN_EXP = (-1021)                 # min binary exponent
_DBL_RADIX = 2                       # exponent radix
DBL_TRUE_MIN = 4.9406564584124654e-324  # min positive value

FLT_DECIMAL_DIG = 9                       # # of decimal digits of rounding precision
FLT_DIG  = 6                       # # of decimal digits of precision
FLT_EPSILON = 1.192092896e-07        # smallest such that 1.0+FLT_EPSILON != 1.0
FLT_HAS_SUBNORM = 1                       # type does support subnormal numbers
FLT_GUARD = 0
FLT_MANT_DIG = 24                      # # of bits in mantissa
FLT_MAX = 3.402823466e+38        # max value
FLT_MAX_10_EXP = 38                      # max decimal exponent
FLT_MAX_EXP = 128                     # max binary exponent
FLT_MIN = 1.175494351e-38        # min normalized positive value
FLT_MIN_10_EXP = (-37)                   # min decimal exponent
FLT_MIN_EXP = (-125)                  # min binary exponent
FLT_NORMALIZE = 0
FLT_RADIX = 2                       # exponent radix
FLT_TRUE_MIN = 1.401298464e-45        # min positive value

LDBL_DIG = DBL_DIG                 # # of decimal digits of precision
LDBL_EPSILON = DBL_EPSILON             # smallest such that 1.0+LDBL_EPSILON != 1.0
LDBL_HAS_SUBNORM = DBL_HAS_SUBNORM         # type does support subnormal numbers
LDBL_MANT_DIG = DBL_MANT_DIG            # # of bits in mantissa
LDBL_MAX = DBL_MAX                 # max value
LDBL_MAX_10_EXP =  DBL_MAX_10_EXP          # max decimal exponent
LDBL_MAX_EXP = DBL_MAX_EXP             # max binary exponent
LDBL_MIN = DBL_MIN                 # min normalized positive value
LDBL_MIN_10_EXP = DBL_MIN_10_EXP          # min decimal exponent
LDBL_MIN_EXP = DBL_MIN_EXP             # min binary exponent
_LDBL_RADIX = _DBL_RADIX              # exponent radix
LDBL_TRUE_MIN = DBL_TRUE_MIN            # min positive value

DECIMAL_DIG = DBL_DECIMAL_DIG


X3DAUDIO_HANDLE_BYTESIZE = 20
X3DAUDIO_HANDLE = (BYTE * X3DAUDIO_HANDLE_BYTESIZE)


# speed of sound in meters per second for dry air at approximately 20C, used with X3DAudioInitialize
X3DAUDIO_SPEED_OF_SOUND = 343.5


X3DAUDIO_CALCULATE_MATRIX = 0x00000001  # enable matrix coefficient table calculation
X3DAUDIO_CALCULATE_DELAY = 0x00000002  # enable delay time array calculation (stereo final mix only)
X3DAUDIO_CALCULATE_LPF_DIRECT = 0x00000004  # enable LPF direct-path coefficient calculation
X3DAUDIO_CALCULATE_LPF_REVERB = 0x00000008  # enable LPF reverb-path coefficient calculation
X3DAUDIO_CALCULATE_REVERB = 0x00000010  # enable reverb send level calculation
X3DAUDIO_CALCULATE_DOPPLER = 0x00000020  # enable doppler shift factor calculation
X3DAUDIO_CALCULATE_EMITTER_ANGLE = 0x00000040  # enable emitter-to-listener interior angle calculation
X3DAUDIO_CALCULATE_ZEROCENTER = 0x00010000  # do not position to front center speaker, signal positioned to remaining speakers instead, front center destination channel will be zero in returned matrix coefficient table, valid only for matrix calculations with final mix formats that have a front center channel
X3DAUDIO_CALCULATE_REDIRECT_TO_LFE = 0x00020000  # apply equal mix of all source channels to LFE destination channel, valid only for matrix calculations with sources that have no LFE channel and final mix formats that have an LFE channel

default_dsp_calculation = X3DAUDIO_CALCULATE_MATRIX | X3DAUDIO_CALCULATE_DOPPLER

X3DAudioInitialize = x3d_lib.X3DAudioInitialize
X3DAudioInitialize.restype = HRESULT
X3DAudioInitialize.argtypes = [c_int, c_float, c_void_p]


X3DAudioCalculate = x3d_lib.X3DAudioCalculate
X3DAudioCalculate.restype = c_void
X3DAudioCalculate.argtypes = [POINTER(X3DAUDIO_HANDLE), POINTER(X3DAUDIO_LISTENER), POINTER(X3DAUDIO_EMITTER), UINT32, POINTER(X3DAUDIO_DSP_SETTINGS)]


AudioCategory_Other = 0
AudioCategory_ForegroundOnlyMedia = 1
AudioCategory_Communications = 3
AudioCategory_Alerts = 4
AudioCategory_SoundEffects = 5
AudioCategory_GameEffects = 6
AudioCategory_GameMedia = 7
AudioCategory_GameChat = 8
AudioCategory_Speech = 9
AudioCategory_Movie = 10
AudioCategory_Media = 11

# Reverb not implemented but if someone wants to take a stab at it.
class XAUDIO2FX_REVERB_PARAMETERS(Structure):
    _fields_ = [
        ('WetDryMix', c_float),  #  ratio of wet (processed) signal to dry (original) signal

        # Delay times
        ('ReflectionsDelay', UINT32),  #  [0, 300] in ms
        ('ReverbDelay', BYTE),  # [0, 85] in ms
        ('RearDelay', UINT32),  # 7.1: [0, 20] in ms, all other: [0, 5] in ms
        ('SideDelay', UINT32),  # .1: [0, 5] in ms, all other: not used, but still validated  # WIN 10 only.

        # Indexed Paremeters
        ('PositionLeft', BYTE),  # [0, 30] no units
        ('PositionRight', BYTE),  # 0, 30] no units, ignored when configured to mono
        ('PositionMatrixLeft', BYTE),  # [0, 30] no units
        ('PositionMatrixRight', BYTE),  # [0, 30] no units, ignored when configured to mono
        ('EarlyDiffusion', BYTE), # [0, 15] no units
        ('LateDiffusion', BYTE),  # [0, 15] no units
        ('LowEQGain', BYTE),  # [0, 12] no units
        ('LowEQCutoff', BYTE),  # [0, 9] no units
        ('LowEQCutoff', BYTE),  # [0, 8] no units
        ('HighEQCutoff', BYTE),  # [0, 14] no units

        # Direct parameters
        ('RoomFilterFreq', c_float),  # [20, 20000] in Hz
        ('RoomFilterMain', c_float),  # [-100, 0] in dB
        ('RoomFilterHF', c_float),  # [-100, 0] in dB
        ('ReflectionsGain', c_float),  # [-100, 20] in dB
        ('ReverbGain', c_float),  # [-100, 20] in dB
        ('DecayTime', c_float),  # [0.1, inf] in seconds
        ('Density', c_float),  # [0, 100] (percentage)
        ('RoomSize', c_float),  # [1, 100] in feet

        # component control
        ('DisableLateField', c_bool),  # TRUE to disable late field reflections
    ]


class IXAudio2(com.pIUnknown):
    _methods_ = [
        ('RegisterForCallbacks',
           com.STDMETHOD(POINTER(IXAudio2EngineCallback))),
        ('UnregisterForCallbacks',
           com.METHOD(ctypes.c_void_p, POINTER(IXAudio2EngineCallback))),
        ('CreateSourceVoice',
         com.STDMETHOD(POINTER(IXAudio2SourceVoice), POINTER(WAVEFORMATEX), UINT32, c_float,
                       POINTER(IXAudio2VoiceCallback), POINTER(XAUDIO2_VOICE_SENDS), POINTER(XAUDIO2_EFFECT_CHAIN))),
        ('CreateSubmixVoice',
         com.STDMETHOD(POINTER(IXAudio2SubmixVoice), UINT32, UINT32, UINT32, UINT32,
                       POINTER(XAUDIO2_VOICE_SENDS), POINTER(XAUDIO2_EFFECT_CHAIN))),
        ('CreateMasteringVoice',
         com.STDMETHOD(POINTER(IXAudio2MasteringVoice), UINT32, UINT32, UINT32, LPCWSTR, POINTER(XAUDIO2_EFFECT_CHAIN),
                       UINT32)),
        ('StartEngine',
         com.STDMETHOD()),
        ('StopEngine',
         com.STDMETHOD()),
        ('CommitChanges',
         com.STDMETHOD(UINT32)),
        ('GetPerformanceData',
         com.METHOD(c_void, POINTER(XAUDIO2_PERFORMANCE_DATA))),
        ('SetDebugConfiguration',
         com.STDMETHOD(POINTER(XAUDIO2_DEBUG_CONFIGURATION), c_void_p)),
    ]


XAudio2Create = xaudio2_lib.XAudio2Create
XAudio2Create.restype = HRESULT
XAudio2Create.argtypes = [POINTER(IXAudio2), UINT32, UINT32]

CreateAudioReverb = xaudio2_lib.CreateAudioReverb
CreateAudioReverb.restype = HRESULT
CreateAudioReverb.argtypes = [POINTER(com.pIUnknown)]




# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

from pyglet import gl
from pyglet import app
from pyglet import window
from pyglet import canvas


class Display:
    """A display device supporting one or more screens.
    
    .. versionadded:: 1.2
    """

    name = None
    """Name of this display, if applicable.

    :type: str
    """

    x_screen = None
    """The X11 screen number of this display, if applicable.

    :type: int
    """

    def __init__(self, name=None, x_screen=None):
        """Create a display connection for the given name and screen.

        On X11, :attr:`name` is of the form ``"hostname:display"``, where the
        default is usually ``":1"``.  On X11, :attr:`x_screen` gives the X 
        screen number to use with this display.  A pyglet display can only be 
        used with one X screen; open multiple display connections to access
        multiple X screens.  
        
        Note that TwinView, Xinerama, xrandr and other extensions present
        multiple monitors on a single X screen; this is usually the preferred
        mechanism for working with multiple monitors under X11 and allows each
        screen to be accessed through a single pyglet`~pyglet.canvas.Display`

        On platforms other than X11, :attr:`name` and :attr:`x_screen` are 
        ignored; there is only a single display device on these systems.

        :Parameters:
            name : str
                The name of the display to connect to.
            x_screen : int
                The X11 screen number to use.

        """
        canvas._displays.add(self)

    def get_screens(self):
        """Get the available screens.

        A typical multi-monitor workstation comprises one :class:`Display`
        with multiple :class:`Screen` s.  This method returns a list of 
        screens which can be enumerated to select one for full-screen display.

        For the purposes of creating an OpenGL config, the default screen
        will suffice.

        :rtype: list of :class:`Screen`
        """
        raise NotImplementedError('abstract')

    def get_default_screen(self):
        """Get the default screen as specified by the user's operating system
        preferences.

        :rtype: :class:`Screen`
        """
        return self.get_screens()[0]

    def get_windows(self):
        """Get the windows currently attached to this display.

        :rtype: sequence of :class:`~pyglet.window.Window`
        """
        return [window for window in app.windows if window.display is self]


class Screen:
    """A virtual monitor that supports fullscreen windows.

    Screens typically map onto a physical display such as a
    monitor, television or projector.  Selecting a screen for a window
    has no effect unless the window is made fullscreen, in which case
    the window will fill only that particular virtual screen.

    The :attr:`width` and :attr:`height` attributes of a screen give the 
    current resolution of the screen.  The :attr:`x` and :attr:`y` attributes 
    give the global location of the top-left corner of the screen.  This is 
    useful for determining if screens are arranged above or next to one 
    another.
    
    Use :func:`~Display.get_screens` or :func:`~Display.get_default_screen`
    to obtain an instance of this class.
    """

    def __init__(self, display, x, y, width, height):
        """
        
        :parameters:
            `display` : `~pyglet.canvas.Display`
                :attr:`display`
            `x` : int
                Left edge :attr:`x`
            `y` : int
                Top edge :attr:`y`
            `width` : int
                :attr:`width`
            `height` : int
                :attr:`height`
        """
        self.display = display
        """Display this screen belongs to."""
        self.x = x
        """Left edge of the screen on the virtual desktop."""
        self.y = y
        """Top edge of the screen on the virtual desktop."""
        self.width = width
        """Width of the screen, in pixels."""
        self.height = height
        """Height of the screen, in pixels."""

    def __repr__(self):
        return '%s(x=%d, y=%d, width=%d, height=%d)' % \
               (self.__class__.__name__, self.x, self.y, self.width, self.height)

    def get_best_config(self, template=None):
        """Get the best available GL config.

        Any required attributes can be specified in `template`.  If
        no configuration matches the template,
        :class:`~pyglet.window.NoSuchConfigException` will be raised.

        :deprecated: Use :meth:`pyglet.gl.Config.match`.

        :Parameters:
            `template` : `pyglet.gl.Config`
                A configuration with desired attributes filled in.

        :rtype: :class:`~pyglet.gl.Config`
        :return: A configuration supported by the platform that best
            fulfils the needs described by the template.
        """
        configs = None
        if template is None:
            for template_config in [gl.Config(double_buffer=True, depth_size=24),
                                    gl.Config(double_buffer=True, depth_size=16),
                                    None]:
                try:
                    configs = self.get_matching_configs(template_config)
                    break
                except window.NoSuchConfigException:
                    pass
        else:
            configs = self.get_matching_configs(template)
        if not configs:
            raise window.NoSuchConfigException()
        return configs[0]

    def get_matching_configs(self, template):
        """Get a list of configs that match a specification.

        Any attributes specified in `template` will have values equal
        to or greater in each returned config.  If no configs satisfy
        the template, an empty list is returned.

        :deprecated: Use :meth:`pyglet.gl.Config.match`.

        :Parameters:
            `template` : `pyglet.gl.Config`
                A configuration with desired attributes filled in.

        :rtype: list of :class:`~pyglet.gl.Config`
        :return: A list of matching configs.
        """
        raise NotImplementedError('abstract')

    def get_modes(self):
        """Get a list of screen modes supported by this screen.

        :rtype: list of :class:`ScreenMode`

        .. versionadded:: 1.2
        """
        raise NotImplementedError('abstract')

    def get_mode(self):
        """Get the current display mode for this screen.

        :rtype: :class:`ScreenMode`

        .. versionadded:: 1.2
        """
        raise NotImplementedError('abstract')

    def get_closest_mode(self, width, height):
        """Get the screen mode that best matches a given size.

        If no supported mode exactly equals the requested size, a larger one
        is returned; or ``None`` if no mode is large enough.

        :Parameters:
            `width` : int
                Requested screen width.
            `height` : int
                Requested screen height.

        :rtype: :class:`ScreenMode`

        .. versionadded:: 1.2
        """
        # Best mode is one with smallest resolution larger than width/height,
        # with depth and refresh rate equal to current mode.
        current = self.get_mode()

        best = None
        for mode in self.get_modes():
            # Reject resolutions that are too small
            if mode.width < width or mode.height < height:
                continue

            if best is None:
                best = mode

            # Must strictly dominate dimensions
            if (mode.width <= best.width and mode.height <= best.height and
                    (mode.width < best.width or mode.height < best.height)):
                best = mode

            # Preferably match rate, then depth.
            if mode.width == best.width and mode.height == best.height:
                points = 0
                if mode.rate == current.rate:
                    points += 2
                if best.rate == current.rate:
                    points -= 2
                if mode.depth == current.depth:
                    points += 1
                if best.depth == current.depth:
                    points -= 1
                if points > 0:
                    best = mode
        return best

    def set_mode(self, mode):
        """Set the display mode for this screen.

        The mode must be one previously returned by :meth:`get_mode` or 
        :meth:`get_modes`.

        :Parameters:
            `mode` : `ScreenMode`
                Screen mode to switch this screen to.

        """
        raise NotImplementedError('abstract')

    def restore_mode(self):
        """Restore the screen mode to the user's default.
        """
        raise NotImplementedError('abstract')


class ScreenMode:
    """Screen resolution and display settings.

    Applications should not construct `ScreenMode` instances themselves; see
    :meth:`Screen.get_modes`.

    The :attr:`depth` and :attr:`rate` variables may be ``None`` if the 
    operating system does not provide relevant data.

    .. versionadded:: 1.2

    """

    width = None
    """Width of screen, in pixels.

    :type: int
    """
    height = None
    """Height of screen, in pixels.

    :type: int
    """
    depth = None
    """Pixel color depth, in bits per pixel.

    :type: int
    """
    rate = None
    """Screen refresh rate in Hz.

    :type: int
    """

    def __init__(self, screen):
        """
        
        :parameters:
            `screen` : `Screen`
        """
        self.screen = screen

    def __repr__(self):
        return '%s(width=%r, height=%r, depth=%r, rate=%r)' % (
            self.__class__.__name__,
            self.width, self.height, self.depth, self.rate)


class Canvas:
    """Abstract drawing area.

    Canvases are used internally by pyglet to represent drawing areas --
    either within a window or full-screen.

    .. versionadded:: 1.2
    """

    def __init__(self, display):
        """
        
        :parameters:
            `display` : `Display`
                :attr:`display`
                
        """
        self.display = display
        """Display this canvas was created on."""



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

"""Memory allocation algorithm for vertex arrays and buffers.

The region allocator is used to allocate vertex indices within a vertex
domain's  multiple buffers.  ("Buffer" refers to any abstract buffer presented
by :py:mod:`pyglet.graphics.vertexbuffer`.
 
The allocator will at times request more space from the buffers. The current
policy is to double the buffer size when there is not enough room to fulfil an
allocation.  The buffer is never resized smaller.

The allocator maintains references to free space only; it is the caller's
responsibility to maintain the allocated regions.
"""

# Common cases:
# -regions will be the same size (instances of same object, e.g. sprites)
# -regions will not usually be resized (only exception is text)
# -alignment of 4 vertices (glyphs, sprites, images, ...)
#
# Optimise for:
# -keeping regions adjacent, reduce the number of entries in glMultiDrawArrays
# -finding large blocks of allocated regions quickly (for drawing)
# -finding block of unallocated space is the _uncommon_ case!
#
# Decisions:
# -don't over-allocate regions to any alignment -- this would require more
#  work in finding the allocated spaces (for drawing) and would result in
#  more entries in glMultiDrawArrays
# -don't move blocks when they truncate themselves.  try not to allocate the
#  space they freed too soon (they will likely need grow back into it later,
#  and growing will usually require a reallocation).
# -allocator does not track individual allocated regions.  Trusts caller
#  to provide accurate (start, size) tuple, which completely describes
#  a region from the allocator's point of view.
# -this means that compacting is probably not feasible, or would be hideously
#  expensive


class AllocatorMemoryException(Exception):
    """The buffer is not large enough to fulfil an allocation.

    Raised by `Allocator` methods when the operation failed due to
    lack of buffer space.  The buffer should be increased to at least
    requested_capacity and then the operation retried (guaranteed to
    pass second time).
    """

    def __init__(self, requested_capacity):
        self.requested_capacity = requested_capacity


class Allocator:
    """Buffer space allocation implementation."""

    __slots__ = 'capacity', 'starts', 'sizes'

    def __init__(self, capacity):
        """Create an allocator for a buffer of the specified capacity.

        :Parameters:
            `capacity` : int
                Maximum size of the buffer.

        """
        self.capacity = capacity

        # Allocated blocks.  Start index and size in parallel lists.
        #
        # # = allocated, - = free
        #
        #  0  3 5        15   20  24                    40
        # |###--##########-----####----------------------|
        #
        # starts = [0, 5, 20]
        # sizes = [3, 10, 4]
        #
        # To calculate free blocks:
        # for i in range(0, len(starts)):
        #   free_start[i] = starts[i] + sizes[i]
        #   free_size[i] =  starts[i+1] - free_start[i]
        # free_size[i+1] = self.capacity - free_start[-1]

        self.starts = []
        self.sizes = []

    def set_capacity(self, size):
        """Resize the maximum buffer size.
        
        The capaity cannot be reduced.

        :Parameters:
            `size` : int
                New maximum size of the buffer.

        """
        assert size > self.capacity
        self.capacity = size

    def alloc(self, size):
        """Allocate memory in the buffer.

        Raises `AllocatorMemoryException` if the allocation cannot be
        fulfilled.

        :Parameters:
            `size` : int
                Size of region to allocate.
               
        :rtype: int
        :return: Starting index of the allocated region.
        """
        assert size >= 0

        if size == 0:
            return 0

        # Return start, or raise AllocatorMemoryException
        if not self.starts:
            if size <= self.capacity:
                self.starts.append(0)
                self.sizes.append(size)
                return 0
            else:
                raise AllocatorMemoryException(size)

        # Restart from zero if space exists
        if self.starts[0] > size:
            self.starts.insert(0, 0)
            self.sizes.insert(0, size)
            return 0

        # Allocate in a free space
        free_start = self.starts[0] + self.sizes[0]
        for i, (alloc_start, alloc_size) in enumerate(zip(self.starts[1:], self.sizes[1:])):
            # Danger!  
            # i is actually index - 1 because of slicing above...
            # starts[i]   points to the block before this free space
            # starts[i+1] points to the block after this free space, and is always valid.
            free_size = alloc_start - free_start
            if free_size == size:
                # Merge previous block with this one (removing this free space)
                self.sizes[i] += free_size + alloc_size
                del self.starts[i+1]
                del self.sizes[i+1]
                return free_start
            elif free_size > size:
                # Increase size of previous block to intrude into this free
                # space.
                self.sizes[i] += size
                return free_start
            free_start = alloc_start + alloc_size

        # Allocate at end of capacity
        free_size = self.capacity - free_start
        if free_size >= size:
            self.sizes[-1] += size
            return free_start
        
        raise AllocatorMemoryException(self.capacity + size - free_size)

    def realloc(self, start, size, new_size):
        """Reallocate a region of the buffer.

        This is more efficient than separate `dealloc` and `alloc` calls, as
        the region can often be resized in-place.

        Raises `AllocatorMemoryException` if the allocation cannot be
        fulfilled.

        :Parameters:
            `start` : int
                Current starting index of the region.
            `size` : int
                Current size of the region.
            `new_size` : int
                New size of the region.

        """
        assert size >= 0 and new_size >= 0
        
        if new_size == 0:
            if size != 0:
                self.dealloc(start, size)
            return 0
        elif size == 0:
            return self.alloc(new_size)

        # return start, or raise AllocatorMemoryException

        # Truncation is the same as deallocating the tail cruft
        if new_size < size:
            self.dealloc(start + new_size, size - new_size)
            return start
            
        # Find which block it lives in
        for i, (alloc_start, alloc_size) in enumerate(zip(*(self.starts, self.sizes))):
            p = start - alloc_start
            if p >= 0 and size <= alloc_size - p:
                break
        if not (p >= 0 and size <= alloc_size - p):
            print(list(zip(self.starts, self.sizes)))
            print(start, size, new_size)
            print(p, alloc_start, alloc_size)
        assert p >= 0 and size <= alloc_size - p, 'Region not allocated'

        if size == alloc_size - p:
            # Region is at end of block. Find how much free space is after it.
            is_final_block = i == len(self.starts) - 1
            if not is_final_block:
                free_size = self.starts[i + 1] - (start + size)
            else:
                free_size = self.capacity - (start + size)

            # TODO If region is an entire block being an island in free space, 
            # can possibly extend in both directions.

            if free_size == new_size - size and not is_final_block:
                # Merge block with next (region is expanded in place to
                # exactly fill the free space)
                self.sizes[i] += free_size + self.sizes[i + 1]
                del self.starts[i + 1]
                del self.sizes[i + 1]
                return start
            elif free_size > new_size - size:
                # Expand region in place
                self.sizes[i] += new_size - size
                return start

        # The block must be repositioned.  Dealloc then alloc.
        
        # But don't do this!  If alloc fails, we've already silently dealloc'd
        # the original block.
        #   self.dealloc(start, size)
        #   return self.alloc(new_size)

        # It must be alloc'd first.  We're not missing an optimisation 
        # here, because if freeing the block would've allowed for the block to 
        # be placed in the resulting free space, one of the above in-place
        # checks would've found it.
        result = self.alloc(new_size)
        self.dealloc(start, size)
        return result

    def dealloc(self, start, size):
        """Free a region of the buffer.

        :Parameters:
            `start` : int
                Starting index of the region.
            `size` : int
                Size of the region.

        """
        assert size >= 0

        if size == 0:
            return

        assert self.starts
        
        # Find which block needs to be split
        for i, (alloc_start, alloc_size) in enumerate(zip(*(self.starts, self.sizes))):
            p = start - alloc_start
            if p >= 0 and size <= alloc_size - p:
                break
        
        # Assert we left via the break
        assert p >= 0 and size <= alloc_size - p, 'Region not allocated'

        if p == 0 and size == alloc_size:
            # Remove entire block
            del self.starts[i]
            del self.sizes[i]
        elif p == 0:
            # Truncate beginning of block
            self.starts[i] += size
            self.sizes[i] -= size
        elif size == alloc_size - p:
            # Truncate end of block
            self.sizes[i] -= size
        else:
            # Reduce size of left side, insert block at right side
            #   $ = dealloc'd block, # = alloc'd region from same block
            #
            #   <------8------>
            #   <-5-><-6-><-7->
            #   1    2    3    4
            #   #####$$$$$#####
            #
            #   1 = alloc_start
            #   2 = start
            #   3 = start + size
            #   4 = alloc_start + alloc_size
            #   5 = start - alloc_start = p
            #   6 = size
            #   7 = {8} - ({5} + {6}) = alloc_size - (p + size)
            #   8 = alloc_size
            #
            self.sizes[i] = p
            self.starts.insert(i + 1, start + size)
            self.sizes.insert(i + 1, alloc_size - (p + size))

    def get_allocated_regions(self):
        """Get a list of (aggregate) allocated regions.

        The result of this method is ``(starts, sizes)``, where ``starts`` is
        a list of starting indices of the regions and ``sizes`` their
        corresponding lengths.

        :rtype: (list, list)
        """
        # return (starts, sizes); len(starts) == len(sizes)
        return self.starts, self.sizes

    def get_fragmented_free_size(self):
        """Returns the amount of space unused, not including the final
        free block.

        :rtype: int
        """
        if not self.starts:
            return 0

        # Variation of search for free block.
        total_free = 0
        free_start = self.starts[0] + self.sizes[0]
        for i, (alloc_start, alloc_size) in enumerate(zip(self.starts[1:], self.sizes[1:])):
            total_free += alloc_start - free_start
            free_start = alloc_start + alloc_size

        return total_free

    def get_free_size(self):
        """Return the amount of space unused.
        
        :rtype: int
        """
        if not self.starts:
            return self.capacity

        free_end = self.capacity - (self.starts[-1] + self.sizes[-1])
        return self.get_fragmented_free_size() + free_end

    def get_usage(self):
        """Return fraction of capacity currently allocated.
        
        :rtype: float
        """
        return 1. - self.get_free_size() / float(self.capacity)

    def get_fragmentation(self):
        """Return fraction of free space that is not expandable.
        
        :rtype: float
        """
        free_size = self.get_free_size()
        if free_size == 0:
            return 0.
        return self.get_fragmented_free_size() / float(self.get_free_size())

    def __str__(self):
        return 'allocs=' + repr(list(zip(self.starts, self.sizes)))

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, str(self))
